from ..models import State, Chamber, District, simple_numbered_districts

NH = State(
    name="New Hampshire",
    abbr="NH",
    capital="Concord",
    capital_tz="America/New_York",
    fips="33",
    unicameral=False,
    legislature_name="New Hampshire General Court",
    legislature_organization_id="ocd-organization/d8dce079-d59d-407c-be93-9d3fce72bd48",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/59217435-34d3-59ec-9673-733ae8b276d3",
    division_id="ocd-division/country:us/state:nh",
    jurisdiction_id="ocd-jurisdiction/country:us/state:nh/government",
    url="http://gencourt.state.nh.us",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/3b857974-4029-4508-94d4-ce3b86c64602",
        num_seats=400,
        title="Representative",
        districts=[
            District(
                "Belknap 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_1",
                1,
            ),
            District(
                "Belknap 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_2",
                2,
            ),
            District(
                "Belknap 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_3",
                1,
            ),
            District(
                "Belknap 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_4",
                1,
            ),
            District(
                "Belknap 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_5",
                4,
            ),
            District(
                "Belknap 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_6",
                4,
            ),
            District(
                "Belknap 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_7",
                3,
            ),
            District(
                "Belknap 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:belknap_8",
                2,
            ),
            District(
                "Carroll 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_1",
                3,
            ),
            District(
                "Carroll 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_2",
                2,
            ),
            District(
                "Carroll 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_3",
                2,
            ),
            District(
                "Carroll 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_4",
                2,
            ),
            District(
                "Carroll 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_5",
                1,
            ),
            District(
                "Carroll 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_6",
                2,
            ),
            District(
                "Carroll 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_7",
                1,
            ),
            District(
                "Carroll 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:carroll_8",
                2,
            ),
            District(
                "Cheshire 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_1",
                1,
            ),
            District(
                "Cheshire 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_2",
                1,
            ),
            District(
                "Cheshire 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_3",
                1,
            ),
            District(
                "Cheshire 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_4",
                1,
            ),
            District(
                "Cheshire 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_5",
                1,
            ),
            District(
                "Cheshire 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_6",
                2,
            ),
            District(
                "Cheshire 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_7",
                1,
            ),
            District(
                "Cheshire 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_8",
                1,
            ),
            District(
                "Cheshire 9",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_9",
                1,
            ),
            District(
                "Cheshire 10",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_10",
                2,
            ),
            District(
                "Cheshire 11",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_11",
                1,
            ),
            District(
                "Cheshire 12",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_12",
                1,
            ),
            District(
                "Cheshire 13",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_13",
                1,
            ),
            District(
                "Cheshire 14",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_14",
                1,
            ),
            District(
                "Cheshire 15",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_15",
                2,
            ),
            District(
                "Cheshire 16",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_16",
                1,
            ),
            District(
                "Cheshire 17",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_17",
                1,
            ),
            District(
                "Cheshire 18",
                "lower",
                "ocd-division/country:us/state:nh/sldl:cheshire_18",
                2,
            ),
            District(
                "Coos 1", "lower", "ocd-division/country:us/state:nh/sldl:coos_1", 2
            ),
            District(
                "Coos 2", "lower", "ocd-division/country:us/state:nh/sldl:coos_2", 1
            ),
            District(
                "Coos 3", "lower", "ocd-division/country:us/state:nh/sldl:coos_3", 1
            ),
            District(
                "Coos 4", "lower", "ocd-division/country:us/state:nh/sldl:coos_4", 1
            ),
            District(
                "Coos 5", "lower", "ocd-division/country:us/state:nh/sldl:coos_5", 2
            ),
            District(
                "Coos 6", "lower", "ocd-division/country:us/state:nh/sldl:coos_6", 1
            ),
            District(
                "Coos 7", "lower", "ocd-division/country:us/state:nh/sldl:coos_7", 1
            ),
            District(
                "Grafton 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_1",
                3,
            ),
            District(
                "Grafton 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_2",
                1,
            ),
            District(
                "Grafton 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_3",
                1,
            ),
            District(
                "Grafton 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_4",
                1,
            ),
            District(
                "Grafton 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_5",
                2,
            ),
            District(
                "Grafton 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_6",
                1,
            ),
            District(
                "Grafton 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_7",
                1,
            ),
            District(
                "Grafton 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_8",
                3,
            ),
            District(
                "Grafton 9",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_9",
                1,
            ),
            District(
                "Grafton 10",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_10",
                1,
            ),
            District(
                "Grafton 11",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_11",
                1,
            ),
            District(
                "Grafton 12",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_12",
                4,
            ),
            District(
                "Grafton 13",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_13",
                1,
            ),
            District(
                "Grafton 14",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_14",
                1,
            ),
            District(
                "Grafton 15",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_15",
                1,
            ),
            District(
                "Grafton 16",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_16",
                1,
            ),
            District(
                "Grafton 17",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_17",
                1,
            ),
            District(
                "Grafton 18",
                "lower",
                "ocd-division/country:us/state:nh/sldl:grafton_18",
                1,
            ),
            District(
                "Hillsborough 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_1",
                4,
            ),
            District(
                "Hillsborough 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_2",
                7,
            ),
            District(
                "Hillsborough 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_3",
                3,
            ),
            District(
                "Hillsborough 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_4",
                3,
            ),
            District(
                "Hillsborough 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_5",
                3,
            ),
            District(
                "Hillsborough 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_6",
                3,
            ),
            District(
                "Hillsborough 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_7",
                3,
            ),
            District(
                "Hillsborough 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_8",
                3,
            ),
            District(
                "Hillsborough 9",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_9",
                3,
            ),
            District(
                "Hillsborough 10",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_10",
                3,
            ),
            District(
                "Hillsborough 11",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_11",
                3,
            ),
            District(
                "Hillsborough 12",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_12",
                8,
            ),
            District(
                "Hillsborough 13",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_13",
                6,
            ),
            District(
                "Hillsborough 14",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_14",
                2,
            ),
            District(
                "Hillsborough 15",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_15",
                2,
            ),
            District(
                "Hillsborough 16",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_16",
                2,
            ),
            District(
                "Hillsborough 17",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_17",
                2,
            ),
            District(
                "Hillsborough 18",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_18",
                2,
            ),
            District(
                "Hillsborough 19",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_19",
                2,
            ),
            District(
                "Hillsborough 20",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_20",
                2,
            ),
            District(
                "Hillsborough 21",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_21",
                2,
            ),
            District(
                "Hillsborough 22",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_22",
                2,
            ),
            District(
                "Hillsborough 23",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_23",
                2,
            ),
            District(
                "Hillsborough 24",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_24",
                2,
            ),
            District(
                "Hillsborough 25",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_25",
                2,
            ),
            District(
                "Hillsborough 26",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_26",
                2,
            ),
            District(
                "Hillsborough 27",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_27",
                1,
            ),
            District(
                "Hillsborough 28",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_28",
                2,
            ),
            District(
                "Hillsborough 29",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_29",
                4,
            ),
            District(
                "Hillsborough 30",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_30",
                3,
            ),
            District(
                "Hillsborough 31",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_31",
                1,
            ),
            District(
                "Hillsborough 32",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_32",
                3,
            ),
            District(
                "Hillsborough 33",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_33",
                2,
            ),
            District(
                "Hillsborough 34",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_34",
                3,
            ),
            District(
                "Hillsborough 35",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_35",
                2,
            ),
            District(
                "Hillsborough 36",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_36",
                2,
            ),
            District(
                "Hillsborough 37",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_37",
                1,
            ),
            District(
                "Hillsborough 38",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_38",
                2,
            ),
            District(
                "Hillsborough 39",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_39",
                2,
            ),
            District(
                "Hillsborough 40",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_40",
                4,
            ),
            District(
                "Hillsborough 41",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_41",
                3,
            ),
            District(
                "Hillsborough 42",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_42",
                3,
            ),
            District(
                "Hillsborough 43",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_43",
                4,
            ),
            District(
                "Hillsborough 44",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_44",
                2,
            ),
            District(
                "Hillsborough 45",
                "lower",
                "ocd-division/country:us/state:nh/sldl:hillsborough_45",
                1,
            ),
            District(
                "Merrimack 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_1",
                1,
            ),
            District(
                "Merrimack 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_2",
                1,
            ),
            District(
                "Merrimack 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_3",
                2,
            ),
            District(
                "Merrimack 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_4",
                2,
            ),
            District(
                "Merrimack 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_5",
                2,
            ),
            District(
                "Merrimack 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_6",
                1,
            ),
            District(
                "Merrimack 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_7",
                2,
            ),
            District(
                "Merrimack 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_8",
                3,
            ),
            District(
                "Merrimack 9",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_9",
                4,
            ),
            District(
                "Merrimack 10",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_10",
                4,
            ),
            District(
                "Merrimack 11",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_11",
                1,
            ),
            District(
                "Merrimack 12",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_12",
                2,
            ),
            District(
                "Merrimack 13",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_13",
                2,
            ),
            District(
                "Merrimack 14",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_14",
                1,
            ),
            District(
                "Merrimack 15",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_15",
                1,
            ),
            District(
                "Merrimack 16",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_16",
                1,
            ),
            District(
                "Merrimack 17",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_17",
                1,
            ),
            District(
                "Merrimack 18",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_18",
                1,
            ),
            District(
                "Merrimack 19",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_19",
                1,
            ),
            District(
                "Merrimack 20",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_20",
                1,
            ),
            District(
                "Merrimack 21",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_21",
                1,
            ),
            District(
                "Merrimack 22",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_22",
                1,
            ),
            District(
                "Merrimack 23",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_23",
                1,
            ),
            District(
                "Merrimack 24",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_24",
                1,
            ),
            District(
                "Merrimack 25",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_25",
                1,
            ),
            District(
                "Merrimack 26",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_26",
                1,
            ),
            District(
                "Merrimack 27",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_27",
                2,
            ),
            District(
                "Merrimack 28",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_28",
                1,
            ),
            District(
                "Merrimack 29",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_29",
                1,
            ),
            District(
                "Merrimack 30",
                "lower",
                "ocd-division/country:us/state:nh/sldl:merrimack_30",
                1,
            ),
            District(
                "Rockingham 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_1",
                3,
            ),
            District(
                "Rockingham 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_2",
                3,
            ),
            District(
                "Rockingham 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_3",
                1,
            ),
            District(
                "Rockingham 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_4",
                3,
            ),
            District(
                "Rockingham 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_5",
                2,
            ),
            District(
                "Rockingham 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_6",
                1,
            ),
            District(
                "Rockingham 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_7",
                1,
            ),
            District(
                "Rockingham 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_8",
                1,
            ),
            District(
                "Rockingham 9",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_9",
                2,
            ),
            District(
                "Rockingham 10",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_10",
                3,
            ),
            District(
                "Rockingham 11",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_11",
                4,
            ),
            District(
                "Rockingham 12",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_12",
                2,
            ),
            District(
                "Rockingham 13",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_13",
                10,
            ),
            District(
                "Rockingham 14",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_14",
                2,
            ),
            District(
                "Rockingham 15",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_15",
                2,
            ),
            District(
                "Rockingham 16",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_16",
                7,
            ),
            District(
                "Rockingham 17",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_17",
                4,
            ),
            District(
                "Rockingham 18",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_18",
                2,
            ),
            District(
                "Rockingham 19",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_19",
                1,
            ),
            District(
                "Rockingham 20",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_20",
                3,
            ),
            District(
                "Rockingham 21",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_21",
                1,
            ),
            District(
                "Rockingham 22",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_22",
                1,
            ),
            District(
                "Rockingham 23",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_23",
                1,
            ),
            District(
                "Rockingham 24",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_24",
                2,
            ),
            District(
                "Rockingham 25",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_25",
                9,
            ),
            District(
                "Rockingham 26",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_26",
                1,
            ),
            District(
                "Rockingham 27",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_27",
                1,
            ),
            District(
                "Rockingham 28",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_28",
                1,
            ),
            District(
                "Rockingham 29",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_29",
                4,
            ),
            District(
                "Rockingham 30",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_30",
                2,
            ),
            District(
                "Rockingham 31",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_31",
                2,
            ),
            District(
                "Rockingham 32",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_32",
                1,
            ),
            District(
                "Rockingham 33",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_33",
                1,
            ),
            District(
                "Rockingham 34",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_34",
                1,
            ),
            District(
                "Rockingham 35",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_35",
                1,
            ),
            District(
                "Rockingham 36",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_36",
                1,
            ),
            District(
                "Rockingham 37",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_37",
                1,
            ),
            District(
                "Rockingham 38",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_38",
                1,
            ),
            District(
                "Rockingham 39",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_39",
                1,
            ),
            District(
                "Rockingham 40",
                "lower",
                "ocd-division/country:us/state:nh/sldl:rockingham_40",
                1,
            ),
            District(
                "Strafford 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_1",
                2,
            ),
            District(
                "Strafford 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_2",
                3,
            ),
            District(
                "Strafford 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_3",
                1,
            ),
            District(
                "Strafford 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_4",
                3,
            ),
            District(
                "Strafford 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_5",
                1,
            ),
            District(
                "Strafford 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_6",
                1,
            ),
            District(
                "Strafford 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_7",
                1,
            ),
            District(
                "Strafford 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_8",
                1,
            ),
            District(
                "Strafford 9",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_9",
                1,
            ),
            District(
                "Strafford 10",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_10",
                4,
            ),
            District(
                "Strafford 11",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_11",
                3,
            ),
            District(
                "Strafford 12",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_12",
                4,
            ),
            District(
                "Strafford 13",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_13",
                1,
            ),
            District(
                "Strafford 14",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_14",
                1,
            ),
            District(
                "Strafford 15",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_15",
                1,
            ),
            District(
                "Strafford 16",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_16",
                1,
            ),
            District(
                "Strafford 17",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_17",
                1,
            ),
            District(
                "Strafford 18",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_18",
                1,
            ),
            District(
                "Strafford 19",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_19",
                3,
            ),
            District(
                "Strafford 20",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_20",
                1,
            ),
            District(
                "Strafford 21",
                "lower",
                "ocd-division/country:us/state:nh/sldl:strafford_21",
                3,
            ),
            District(
                "Sullivan 1",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_1",
                1,
            ),
            District(
                "Sullivan 2",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_2",
                1,
            ),
            District(
                "Sullivan 3",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_3",
                3,
            ),
            District(
                "Sullivan 4",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_4",
                1,
            ),
            District(
                "Sullivan 5",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_5",
                1,
            ),
            District(
                "Sullivan 6",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_6",
                3,
            ),
            District(
                "Sullivan 7",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_7",
                1,
            ),
            District(
                "Sullivan 8",
                "lower",
                "ocd-division/country:us/state:nh/sldl:sullivan_8",
                2,
            ),
        ],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c95932a0-2ac1-4f14-a0c4-ade8568daac6",
        num_seats=24,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nh", "upper", 24
        ),
    ),
)



from ..models import State, Chamber, simple_numbered_districts

SC = State(
    name="South Carolina",
    abbr="SC",
    capital="Columbia",
    capital_tz="America/New_York",
    fips="45",
    unicameral=False,
    legislature_name="South Carolina Legislature",
    legislature_organization_id="ocd-organization/57920b9f-163e-4cc3-9642-e616579b47ee",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/852b2773-ace9-5d16-b69b-ecc11f74b548",
    division_id="ocd-division/country:us/state:sc",
    jurisdiction_id="ocd-jurisdiction/country:us/state:sc/government",
    url="http://www.scstatehouse.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/11144d16-6b61-4e2a-94fe-6f9c24e36193",
        num_seats=124,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:sc", "lower", 124
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/705ec497-4b18-4a39-a0e0-ce114a2f3c78",
        num_seats=46,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:sc", "upper", 46
        ),
    ),
)



import typing


class OpenStatesError(Exception):
    """Base class for exceptions from core"""


class InternalError(OpenStatesError):
    """Indication something went wrong inside the backend that never should happen"""


class CommandError(OpenStatesError):
    """Errors from within CLI"""


# import-related errors


class DataImportError(OpenStatesError):
    """A generic error related to the import process."""


class InvalidVoteEventError(DataImportError):
    """Attempt to create a vote event without an identifier or bill_id"""


class DuplicateItemError(DataImportError):
    """Attempt was made to import items that resolve to the same database item."""

    def __init__(self, data: dict, obj: typing.Any, data_sources: list[dict] = None):
        # obj.sources can be a list or subobject
        obj_sources = getattr(obj, "sources", [])
        if not isinstance(obj_sources, list):
            obj_sources = list(obj_sources.values_list("url", flat=True))
        super(DuplicateItemError, self).__init__(
            "attempt to import data that would conflict with "
            "data already in the import: {} "
            "(already imported as {})\n"
            "obj1 sources: {}\nobj2 sources: {}".format(
                data,
                obj,
                obj_sources,
                [s["url"] for s in data_sources or []],
            )
        )


class UnresolvedIdError(DataImportError):
    """Attempt was made to resolve an id that has no result."""


# scrape-related errors


class ScrapeError(OpenStatesError):
    """A generic error related to the scrape process."""


class ScrapeValueError(OpenStatesError, ValueError):
    """An invalid value was passed to a scrape object."""


class EmptyScrape(OpenStatesError):
    """Indicate an intentionally empty scrape."""



import pytest  # type: ignore
from openstates.data.models import Jurisdiction, Division, Organization, Post
from openstates.cli.initdb import (
    create_division,
    create_chamber,
    create_full_jurisdiction,
)
from openstates.metadata import lookup
from django.db.utils import IntegrityError  # type: ignore


@pytest.mark.django_db
def test_create_division_basic():
    div = create_division("ocd-division/country:us/state:nc", "North Carolina", "NC")
    assert div.name == "North Carolina"
    assert Division.objects.count() == 1


@pytest.mark.django_db
def test_create_division_duplicate():
    div = create_division("ocd-division/country:us/state:nc", "North Carolina", "NC")
    # first name persists b/c of get_or_create
    div = create_division("ocd-division/country:us/state:nc", "N. Carolina", "NC")
    assert div.name == "North Carolina"
    assert Division.objects.count() == 1


@pytest.mark.django_db
def test_create_chamber_basic():
    nc = lookup(abbr="NC")

    juris = Jurisdiction.objects.create(
        id=nc.jurisdiction_id, name=nc.name, division=None
    )
    leg = Organization.objects.create(
        id=nc.legislature_organization_id,
        name=nc.legislature_name,
        classification="legislature",
        jurisdiction=juris,
    )
    create_chamber(juris, leg, nc.lower, "NC")

    # ensure the org and posts were created
    org = Organization.objects.get(classification="lower")
    assert org.name == nc.lower.name
    assert org.id == nc.lower.organization_id
    assert org.posts.count() == 120


@pytest.mark.django_db
def test_create_chamber_duplicate_idempotent():
    nc = lookup(abbr="NC")

    juris = Jurisdiction.objects.create(
        id=nc.jurisdiction_id, name=nc.name, division=None
    )
    leg = Organization.objects.create(
        id=nc.legislature_organization_id,
        name=nc.legislature_name,
        classification="legislature",
        jurisdiction=juris,
    )

    # second call, identical to first, should be idempotent
    create_chamber(juris, leg, nc.lower, "NC")
    create_chamber(juris, leg, nc.lower, "NC")

    assert Organization.objects.filter(classification="lower").count() == 1

    # ensure the org and posts were created
    org = Organization.objects.get(classification="lower")
    assert org.name == nc.lower.name
    assert org.id == nc.lower.organization_id
    assert org.posts.count() == 120


@pytest.mark.django_db
def test_create_chamber_duplicate_with_changes():
    nc = lookup(abbr="NC")

    juris = Jurisdiction.objects.create(
        id=nc.jurisdiction_id, name=nc.name, division=None
    )
    leg = Organization.objects.create(
        id=nc.legislature_organization_id,
        name=nc.legislature_name,
        classification="legislature",
        jurisdiction=juris,
    )

    create_chamber(juris, leg, nc.lower, "NC")
    # second call, but lower chamber name has been changed
    nc.lower.name = "Ronald McDonald House of Clowns"
    with pytest.raises(IntegrityError):
        create_chamber(juris, leg, nc.lower, "NC")  # unsupported, should definitely be loud

    assert Organization.objects.filter(classification="lower").count() == 1


@pytest.mark.django_db
def test_create_chamber_unicam():
    ne = lookup(abbr="NE")

    juris = Jurisdiction.objects.create(
        id=ne.jurisdiction_id, name=ne.name, division=None
    )
    leg = Organization.objects.create(
        id=ne.legislature_organization_id,
        name=ne.legislature_name,
        classification="legislature",
        jurisdiction=juris,
    )

    create_chamber(juris, leg, ne.legislature, "NE")

    # no org was created, but posts were
    assert Organization.objects.count() == 1
    org = Organization.objects.get(classification="legislature")
    assert org.posts.count() == 49


@pytest.mark.django_db
def test_create_full_jurisdiction_basic():
    nc = lookup(abbr="NC")
    create_full_jurisdiction(nc)

    assert Jurisdiction.objects.count() == 1
    juris = Jurisdiction.objects.get()
    assert juris.name == nc.name
    assert juris.organizations.count() == 4
    assert (
        juris.organizations.get(classification="executive").id
        == nc.executive_organization_id
    )
    assert (
        juris.organizations.get(classification="legislature").id
        == nc.legislature_organization_id
    )
    # 120 + 50
    assert Post.objects.count() == 170


@pytest.mark.django_db
def test_create_full_jurisdiction_idempotent():
    nc = lookup(abbr="NC")
    create_full_jurisdiction(nc)
    # second call, does nothing
    create_full_jurisdiction(nc)

    assert Jurisdiction.objects.count() == 1
    juris = Jurisdiction.objects.get()
    assert juris.name == nc.name
    assert juris.organizations.count() == 4
    assert (
        juris.organizations.get(classification="executive").id
        == nc.executive_organization_id
    )
    assert (
        juris.organizations.get(classification="legislature").id
        == nc.legislature_organization_id
    )
    # 120 + 50
    assert Post.objects.count() == 170


@pytest.mark.django_db
def test_create_federal_jurisdiction():
    us = lookup(abbr="US")
    create_full_jurisdiction(us)

    assert Jurisdiction.objects.count() == 1
    juris = Jurisdiction.objects.get()
    assert juris.name == us.name
    assert juris.classification == "country"
    assert juris.organizations.count() == 4
    assert (
        juris.organizations.get(classification="executive").id
        == us.executive_organization_id
    )
    assert (
        juris.organizations.get(classification="legislature").id
        == us.legislature_organization_id
    )
    # 435 House + 50 Senate posts + 2 territories (DC & PR for now)
    assert Post.objects.filter(role="Representative").count() == 435
    assert Post.objects.filter(role="Senator").count() == 50
    assert Post.objects.filter(role="Delegate").count() == 5
    assert Post.objects.filter(role="Resident Commissioner").count() == 1



from ..models import State, Chamber, simple_numbered_districts

DE = State(
    name="Delaware",
    abbr="DE",
    capital="Dover",
    capital_tz="America/New_York",
    fips="10",
    unicameral=False,
    legislature_name="Delaware General Assembly",
    legislature_organization_id="ocd-organization/d2671846-0bc5-4082-b96d-989b31cc037f",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/dcfaf7fc-52ad-4a49-b88a-3a6fcea19248",
    division_id="ocd-division/country:us/state:de",
    jurisdiction_id="ocd-jurisdiction/country:us/state:de/government",
    url="http://legis.delaware.gov/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/378d45a3-b3f9-41d5-b4f1-6fcf2979666c",
        num_seats=41,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:de", "lower", 41
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/099db7ce-13e0-4865-98ef-9787f65705c4",
        num_seats=21,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:de", "upper", 21
        ),
    ),
)



import typing

from .base import BaseImporter
from ._types import _JsonDict
from ..utils import get_pseudo_id
from ..data.models import (
    Event,
    EventLocation,
    EventDocument,
    EventParticipant,
    EventMedia,
    EventAgendaItem,
    EventRelatedEntity,
    EventAgendaMedia,
)
from .organizations import OrganizationImporter
from .vote_events import VoteEventImporter
from ..metadata import lookup


class EventImporter(BaseImporter):
    _type = "event"
    model_class = Event
    related_models = {
        "documents": (
            EventDocument,
            "event_id",
            {},
        ),
        "participants": (EventParticipant, "event_id", {}),
        "media": (EventMedia, "event_id", {}),
        "agenda": (
            EventAgendaItem,
            "event_id",
            {
                "related_entities": (EventRelatedEntity, "agenda_item_id", {}),
                "media": (EventAgendaMedia, "agenda_item_id", {}),
            },
        ),
    }
    preserve_order = {"agenda"}

    def __init__(
        self,
        jurisdiction_id: str,
        vote_event_importer: VoteEventImporter,
        do_postimport=True,
    ):
        super(EventImporter, self).__init__(jurisdiction_id, do_postimport)
        self.org_importer = OrganizationImporter(jurisdiction_id)
        self.vote_event_importer = vote_event_importer

    def get_object(self, event: _JsonDict) -> Event:
        if event.get("dedupe_key"):
            spec = {
                "dedupe_key": event.get("dedupe_key"),
                "jurisdiction_id": self.jurisdiction_id,
            }
        else:
            spec = {
                "name": event["name"],
                "description": event["description"],
                "start_date": event["start_date"],
                "end_date": event["end_date"],
                "jurisdiction_id": self.jurisdiction_id,
            }
        return self.model_class.objects.get(**spec)

    def get_location(self, location_data: _JsonDict) -> EventLocation:
        obj, created = EventLocation.objects.get_or_create(
            name=location_data["name"],
            url=location_data.get("url", ""),
            jurisdiction_id=self.jurisdiction_id,
        )
        return obj

    def get_chamber_name_from_event_name(
        self, name: typing.Union[str, tuple]
    ) -> typing.Union[str, None]:
        if isinstance(name, tuple):
            event_name = name[0].lower()
        else:
            event_name = name.lower()
        event_name = event_name.split(" ")[0]

        state = lookup(jurisdiction_id=self.jurisdiction_id)
        chamber_types = []
        if state.lower and state.lower.name.lower() in event_name:
            chamber_types.append(state.lower.chamber_type)
        if state.upper and state.upper.name.lower() in event_name:
            chamber_types.append(state.upper.chamber_type)
        if len(chamber_types) == 1:
            return chamber_types[0]
        return None

    def prepare_for_db(self, data: _JsonDict) -> _JsonDict:
        data["jurisdiction_id"] = self.jurisdiction_id
        data["location"] = self.get_location(data["location"])
        org_classification = self.get_chamber_name_from_event_name(data["name"])

        # all objects being inserted should be non-deleted
        data["deleted"] = False
        data["start_date"] = data["start_date"]
        data["end_date"] = data.get("end_date", "")

        for participant in data["participants"]:
            # conversion from old/incorrect names first
            if "committee_id" in participant:
                participant["organization_id"] = participant.pop("committee_id")
            if "legislator_id" in participant:
                participant["person_id"] = participant.pop("legislator_id")

            if "person_id" in participant:
                participant["person_id"] = self.resolve_person(
                    participant["person_id"], org_classification=org_classification
                )
            elif "organization_id" in participant:
                participant["organization_id"] = self.org_importer.resolve_json_id(
                    participant["organization_id"], allow_no_match=True
                )

        for item in data["agenda"]:
            for entity in item["related_entities"]:
                if "person_id" in entity:
                    entity["person_id"] = self.resolve_person(
                        entity["person_id"], org_classification=org_classification
                    )
                elif "organization_id" in entity:
                    entity["organization_id"] = self.org_importer.resolve_json_id(
                        entity["organization_id"], allow_no_match=True
                    )
                elif "bill_id" in entity:
                    bill = get_pseudo_id(entity["bill_id"])
                    entity["bill_id"] = self.resolve_bill(
                        bill["identifier"], date=data["start_date"]
                    )
                elif "vote_event_id" in entity:
                    entity["vote_event_id"] = self.vote_event_importer.resolve_json_id(
                        entity["vote_event_id"], allow_no_match=True
                    )

        return data

    def postimport(self) -> None:
        all_db_ids = self.json_to_db_id.values()
        update_set = Event.objects.filter(
            jurisdiction_id=self.jurisdiction_id, deleted=False
        ).exclude(id__in=all_db_ids)
        update_set.update(deleted=True)



import click
import json
import datetime
import jsonschema
from jsonschema import validate, Draft3Validator
from ..scrape.schemas.bill import schema as bill_schema
from ..scrape.schemas.event import schema as event_schema
from ..scrape.schemas.jurisdiction import schema as jurisdiction_schema
from ..scrape.schemas.organization import schema as organization_schema
from ..scrape.schemas.vote_event import schema as vote_event_schema


@click.command()
@click.argument("scraper_entity_type")
@click.argument("filepath", type=click.Path(exists=True))
def main(
    scraper_entity_type: str,
    filepath: str,
) -> None:
    """Validate a JSON file against scraper data schema. scraper_entity_type may be bill, event, jurisdiction,
    organization, vote_event. filepath needs to be a json file"""
    with open(filepath) as json_file:
        entity_instance = json.load(json_file)

    schema = None
    if scraper_entity_type == "bill":
        schema = bill_schema
    elif scraper_entity_type == "event":
        schema = event_schema
    elif scraper_entity_type == "jurisdiction":
        schema = jurisdiction_schema
    elif scraper_entity_type == "organization":
        schema = organization_schema
    elif scraper_entity_type == "vote_event":
        schema = vote_event_schema

    # See openstates/scrape/base.py for source of this validation code
    type_checker = Draft3Validator.TYPE_CHECKER.redefine(
        "datetime", lambda c, d: isinstance(d, (datetime.date, datetime.datetime))
    )
    type_checker = type_checker.redefine(
        "date",
        lambda c, d: (
            isinstance(d, datetime.date) and not isinstance(d, datetime.datetime)
        ),
    )
    ValidatorCls = jsonschema.validators.extend(
        Draft3Validator, type_checker=type_checker
    )
    validate(instance=entity_instance, schema=schema, cls=ValidatorCls)  # type: ignore


if __name__ == "__main__":
    main()



import typing
from .data import (  # noqa: F401
    STATES_BY_ABBR,
    INTERNATIONAL,
    STATES_BY_JID,
    NON_US_INTERNATIONAL_ABBRS,
    INTERNATIONAL_ABBRS,
    STATES_BY_NAME,
)
from .models import State, Chamber, District


def lookup(
    *,
    abbr: typing.Optional[str] = None,
    jurisdiction_id: typing.Optional[str] = None,
    name: typing.Optional[str] = None,
) -> State:
    if abbr:
        return STATES_BY_ABBR[abbr.upper()]
    if name:
        return STATES_BY_NAME[name.lower()]
    if jurisdiction_id:
        return STATES_BY_JID[jurisdiction_id]
    raise ValueError(f"invalid lookup call {abbr} {jurisdiction_id} {name}")


def lookup_district_with_ancestors(
    *, division_id: str
) -> typing.Tuple[State, Chamber, District]:
    district: typing.Optional[District] = None
    pieces = division_id.split("/")
    if len(pieces) == 4:
        state_portion = pieces[2]
    elif len(pieces) == 3:
        # special case the at-large ids for now, these should be fixed
        if division_id == "ocd-division/country:us/territory:pr":
            state = typing.cast(State, lookup(abbr="pr"))
            chamber = typing.cast(Chamber, state.upper)
        elif division_id == "ocd-division/country:us/district:dc":
            state = typing.cast(State, lookup(abbr="dc"))
            chamber = typing.cast(Chamber, state.legislature)
        district = typing.cast(District, chamber.lookup_district(name="At-Large"))
        return state, chamber, district
    else:
        raise ValueError(f"invalid division id: {division_id}")

    try:
        state = lookup(abbr=state_portion.split(":")[1])
    except KeyError:
        raise ValueError(f"invalid division id: {division_id}, no such state")

    district = state.lookup_district(division_id)

    if not district:
        raise ValueError(f"invalid division id: {division_id}, no such district")

    if state.unicameral:
        chamber = typing.cast(Chamber, state.legislature)
    elif "sldl" in division_id:
        chamber = typing.cast(Chamber, state.lower)
    elif "sldu" in division_id:
        chamber = typing.cast(Chamber, state.upper)

    return state, chamber, district



import re
import typing
import uuid
from collections import defaultdict
import us  # type: ignore
import requests
import click
from ..models.people import (
    Person,
    OtherIdentifier,
    Role,
    Party,
    Office,
    Link,
    PersonIdBlock,
)  # type: ignore
from ..models.committees import Committee, Membership
from ..utils.people import dump_obj, get_data_path

# chosen at random, but needs to be constant
US_UUID_NAMESPACE = uuid.UUID("bf6b57c6-8cfe-454c-bd26-9c2b508c30b2")


def make_person_id(bioguide: str) -> str:
    return "ocd-person/" + str(uuid.uuid5(US_UUID_NAMESPACE, bioguide))


def make_org_id(id_: str) -> str:
    return "ocd-organization/" + str(uuid.uuid5(US_UUID_NAMESPACE, id_))


def sanitize_phone(phone: str) -> str:
    """Remove trail text, toll-free phone number or N/A"""
    if phone.lower() in ["n/a", "same as above"]:
        return ""
    # Some phone might appear like (123) 456 7890
    pattern = r"\((\d{3})\)\s*(\d{3})-(\d{4})"
    match = re.search(pattern, phone)
    if match:
        # Format the first matched number as XXX-XXX-XXXX
        formatted_number = f"{match.group(1)}-{match.group(2)}-{match.group(3)}"
        return formatted_number
    return phone


def _fix_bad_dashes(phone: str) -> str:
    phone = sanitize_phone(phone)
    return phone.replace("–", "-")


def get_district_offices() -> defaultdict[str, list[Office]]:
    district_offices = defaultdict(list)
    url = "https://theunitedstates.io/congress-legislators/legislators-district-offices.json"
    entries = requests.get(url).json()
    for entry in entries:
        for num, office in enumerate(entry["offices"]):
            address = office.get("address", "")
            if address:
                if office.get("suite"):
                    address += " " + office["suite"]
                address += f"; {office['city']}, {office['state']} {office['zip']}"

            district_offices[entry["id"]["bioguide"]].append(
                Office(
                    classification="district",
                    voice=_fix_bad_dashes(office.get("phone", "")),
                    fax=_fix_bad_dashes(office.get("fax", "")),
                    address=address,
                    name=f"District Office #{num+1}",
                )
            )
    return district_offices


def get_social() -> dict[str, PersonIdBlock]:
    social: dict[str, PersonIdBlock] = {}
    url = (
        "https://theunitedstates.io/congress-legislators/legislators-social-media.json"
    )
    entries = requests.get(url).json()
    for entry in entries:
        social[entry["id"]["bioguide"]] = PersonIdBlock(
            twitter=entry["social"].get("twitter", ""),
            facebook=entry["social"].get("facebook", ""),
            youtube=entry["social"].get("youtube_id", ""),
        )
    return social


def fetch_current_people() -> typing.Iterable[tuple[str, Person]]:
    url = "https://theunitedstates.io/congress-legislators/legislators-current.json"
    legislators = requests.get(url).json()
    for leg in legislators:
        yield current_to_person(leg)


def current_to_person(current: dict[str, typing.Any]) -> tuple[str, Person]:
    full_name = current["name"].get(
        "official_full", f"{current['name']['first']} {current['name']['last']}"
    )
    bioguide = current["id"]["bioguide"]
    gender = current["bio"].get("gender", "")
    birthday = current["bio"].get("birthday", "")
    p = Person(
        id=make_person_id(bioguide),
        name=full_name,
        given_name=current["name"]["first"],
        family_name=current["name"]["last"],
        middle_name=current["name"].get("middle", ""),
        gender=gender,
        birth_date=birthday,
        roles=[],
    )
    for key, value in current["id"].items():
        if isinstance(value, list):
            for identifier in value:
                p.other_identifiers.append(
                    OtherIdentifier(scheme=key, identifier=identifier)
                )
        else:
            p.other_identifiers.append(OtherIdentifier(scheme=key, identifier=value))

    # keep mapping of start & end dates of party memberships
    parties: dict[str, dict[str, str]] = defaultdict(dict)
    for term in current["terms"]:
        if term["start"] < parties[term["party"]].get("start", "9999-99-99"):
            parties[term["party"]]["start"] = term["start"]
        if term["end"] > parties[term["party"]].get("end", ""):
            parties[term["party"]]["end"] = term["end"]
        if term["type"] == "sen":
            role_type = "upper"
            district = us.states.lookup(term["state"]).name
            # division_id = f"ocd-division/country:us/state:{term['state']}"
        elif term["type"] == "rep":
            role_type = "lower"
            if term["district"] == 0:
                district = f"{term['state']}-AL"
            else:
                district = f"{term['state']}-{term['district']}"
            # division_id = f"ocd-division/country:us/state:{term['state']}/cd:{term['district']}"
        role = Role(
            type=role_type,
            district=district,
            jurisdiction="ocd-jurisdiction/country:us/government",
            start_date=term["start"],
            end_date=term["end"],
        )
        p.roles.append(role)

    # add party memberships
    for name, times in parties.items():
        if name == "Democrat":
            name = "Democratic"
        p.party.append(
            Party(name=name, start_date=times["start"], end_date=times["end"])
        )

    # add contact info from latest term
    cur_term = current["terms"][-1]
    if "url" in cur_term:
        p.links.append(Link(note="website", url=cur_term["url"]))
    if "contact_form" in cur_term:
        p.links.append(Link(note="contact form", url=cur_term["contact_form"]))

    address = cur_term.get("address", "")
    voice = cur_term.get("phone", "")
    if address or voice:
        p.offices.append(
            Office(
                classification="capitol",
                address=address,
                voice=voice,
            )
        )

    return bioguide, p


def scrape_people() -> None:
    output_dir = get_data_path("us") / "legislature"
    district_offices = get_district_offices()
    social = get_social()
    for bioguide, person in fetch_current_people():
        person.offices.extend(district_offices[bioguide])
        if bioguide in social:
            person.ids = social[bioguide]
        person.sources.append(Link(url="https://theunitedstates.io/"))
        person.image = (
            f"https://theunitedstates.io/images/congress/450x550/{bioguide}.jpg"
        )
        dump_obj(person, output_dir=output_dir)


ThomasMappingType = dict[tuple[str, typing.Optional[str], str], list[str]]


def get_thomas_mapping(convert_chamber: dict) -> ThomasMappingType:
    """
    This function creates a dictionary that maps a tuple to a list of thomas_ids.
    the tuples consist of (chamber, parent, name).
    """
    name_mapping: ThomasMappingType = {}
    url = "https://theunitedstates.io/congress-legislators/committees-current.json"
    committees = requests.get(url).json()

    for com in committees:
        name = com["name"]
        thomas_id = com["thomas_id"]
        chamber = convert_chamber[com["type"]]

        name_mapping[(chamber, None, name)] = name_mapping.get(
            (chamber, None, name), []
        ) + [thomas_id]

        if "subcommittees" in com:
            for sub in com["subcommittees"]:
                sub_name = sub["name"]
                thomas_id_agg = thomas_id + sub["thomas_id"]
                parent_id = make_org_id(thomas_id)
                name_mapping[(chamber, parent_id, sub_name)] = name_mapping.get(
                    (chamber, parent_id, sub_name), []
                ) + [thomas_id_agg]

    return name_mapping


def fetch_current_committees(convert_chamber: dict) -> typing.Iterable[Committee]:
    url = "https://theunitedstates.io/congress-legislators/committees-current.json"
    committees = requests.get(url).json()
    for com in committees:
        committee_name = com["name"]
        thomas_id = com["thomas_id"]
        chamber = convert_chamber[com["type"]]

        c = Committee(
            id=make_org_id(thomas_id),
            jurisdiction="ocd-jurisdiction/country:us/government",
            name=committee_name,
            chamber=chamber,
            members=[Membership(name="placeholder", role="member")],
        )

        if "address" in com:
            c.extras["address"] = com["address"]
        if "phone" in com:
            c.extras["phone"] = com["phone"]
        if "url" in com:
            c.add_link(com["url"], note="homepage")
        if "minority_url" in com:
            c.add_link(com["minority_url"], note="homepage")

        c.extras["type"] = com["type"]

        yield c

        if "subcommittees" in com:
            for sub in com["subcommittees"]:
                subcommittee_name = sub["name"]
                sub_thomas_id = sub["thomas_id"]
                sub_thomas_id = thomas_id + sub_thomas_id
                s = Committee(
                    id=make_org_id(sub_thomas_id),
                    jurisdiction="ocd-jurisdiction/country:us/government",
                    name=subcommittee_name,
                    parent=make_org_id(thomas_id),
                    chamber=chamber,
                    classification="subcommittee",
                    members=[Membership(name="placeholder", role="member")],
                )

                if "address" in sub:
                    s.extras["address"] = sub["address"]
                if "phone" in sub:
                    s.extras["phone"] = sub["phone"]

                s.extras["type"] = com["type"]

                yield s


def get_members_mapping() -> dict[str, list]:
    url = "https://theunitedstates.io/congress-legislators/committee-membership-current.json"
    members_mapping = requests.get(url).json()

    return members_mapping


def grab_members(
    committee: Committee,
    name_mapping: list[str],
    members_mapping: dict[str, list],
) -> None:
    for t_id in name_mapping:
        if t_id in members_mapping:
            members = members_mapping[t_id]
            for member in members:
                if "title" in member:
                    committee.members.append(
                        Membership(
                            name=member["name"],
                            role=member["title"],
                            person_id=make_person_id(member["bioguide"]),
                        )
                    )
                else:
                    committee.members.append(
                        Membership(
                            name=member["name"],
                            role="Member",
                            person_id=make_person_id(member["bioguide"]),
                        )
                    )
    # remove placeholder member
    committee.members.pop(0)


def scrape_committees() -> None:
    output_dir = get_data_path("us") / "committees"

    convert_chamber = {"house": "lower", "senate": "upper", "joint": "legislature"}

    members_mapping = get_members_mapping()
    name_mapping = get_thomas_mapping(convert_chamber)

    for committee in fetch_current_committees(convert_chamber):
        name = committee.name
        grab_members(
            committee,
            name_mapping[(committee.chamber, committee.parent, name)],
            members_mapping,
        )
        committee.sources.append(Link(url="https://theunitedstates.io/"))

        if len(committee.members) > 0:
            dump_obj(committee, output_dir=output_dir)


@click.command()
def main() -> None:
    """
    Create/Update United States legislators from unitedstates.io
    """
    scrape_people()
    scrape_committees()


if __name__ == "__main__":
    main()



# flake8: noqa
from .general import (
    ocd_uuid,
    get_data_path,
    get_all_abbreviations,
    dump_obj,
    get_new_filename,
    legacy_districts,
    load_municipalities,
)
from .retire import retire_file, retire_person
from .images import download_state_images



# Generated by Django 3.0.5 on 2020-04-29 09:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("data", "0008_person_current_state")]

    operations = [
        migrations.AlterField(
            model_name="bill",
            name="first_action_date",
            field=models.CharField(default=None, max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name="bill",
            name="latest_action_date",
            field=models.CharField(default=None, max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name="bill",
            name="latest_action_description",
            field=models.TextField(default=""),
        ),
        migrations.AlterField(
            model_name="bill",
            name="latest_passage_date",
            field=models.CharField(default=None, max_length=25, null=True),
        ),
    ]



from pathlib import Path
from spatula.cli import scrape
from .people import merge as people_merge
from .committees import merge as committees_merge
from ..utils.instrument import Instrumentation
import click
import time

stats = Instrumentation()


@click.command()
@click.argument("abbr")
@click.argument("scraper_type")
@click.option("--scrape-only", help="Only perform scrape.", is_flag=True)
@click.option("--merge-only", help="Only perform merge.", is_flag=True)
@click.option(
    "--fastmode",
    help="Use a cache to avoid making unnecessary requests while developing.",
    is_flag=True,
)
@click.option(
    "--reset-offices",
    is_flag=True,
    help="Reset offices to latest scrape instead of trying to merge old with new.",
)
def main(
    abbr: str,
    scraper_type: str,
    scrape_only: bool,
    merge_only: bool,
    fastmode: bool,
    reset_offices: bool,
) -> None:
    output_dir = Path(f"_scrapes/{abbr}/{scraper_type}")
    if not merge_only:
        stats.write_stats(
            [
                {
                    "metric": "scrapes",
                    "fields": {"total": 1},
                    "tags": {"jurisdiction": abbr, "scraper_type": scraper_type},
                }
            ]
        )
        args = [
            f"scrapers_next.{abbr}.{scraper_type}",
            "-o",
            output_dir,
            "--rmdir",
        ]
        if fastmode:
            args.append("--fastmode")
        try:
            scrape(args)
        except SystemExit as e:
            if e.code != 0:
                raise
        stats.write_stats(
            [
                {
                    "metric": "last_scrape",
                    "fields": {"time": int(time.time())},
                    "tags": {
                        "jurisdiction": abbr,
                        "scraper_type": scraper_type,
                    },
                }
            ]
        )
    if not scrape_only and "people" in scraper_type:
        stats.write_stats(
            [
                {
                    "metric": "people_merges",
                    "fields": {"total": 1},
                    "tags": {"jurisdiction": abbr},
                }
            ]
        )
        merge_args = [abbr, str(output_dir)]
        if reset_offices:
            merge_args.append("--reset-offices")
        people_merge(merge_args)
        stats.write_stats(
            [
                {
                    "metric": "last_people_merge",
                    "fields": {"time": int(time.time())},
                    "tags": {
                        "jurisdiction": abbr,
                    },
                }
            ]
        )
    elif not scrape_only and "committees" in scraper_type:
        merge_args = [abbr, str(output_dir)]
        committees_merge(merge_args)
        stats.write_stats(
            [
                {
                    "metric": "committee_merges",
                    "fields": {"total": 1},
                    "tags": {"jurisdiction": abbr},
                },
                {
                    "metric": "last_committee_merge",
                    "fields": {"time": int(time.time())},
                    "tags": {
                        "jurisdiction": abbr,
                    },
                },
            ]
        )

    stats.close()


if __name__ == "__main__":
    main()



import pytest
import datetime
from openstates.scrape import Event, calculate_window


def event_obj():
    e = Event(
        name="get-together",
        start_date=datetime.datetime.utcnow().isoformat().split(".")[0] + "Z",
        location_name="Joe's Place",
    )
    e.add_source(url="http://example.com/foobar")
    return e


def test_basic_event():
    e = event_obj()
    e.validate()


def test_event_str():
    e = event_obj()
    assert e.name in str(e)


def test_bad_event():
    e = event_obj()
    e.start_date = 6

    with pytest.raises(ValueError):
        e.validate()


def test_basic_agenda():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda["description"] == "foo bar"
    assert e.agenda[0] == agenda
    e.validate()


def test_agenda_add_person():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda["related_entities"] == []

    agenda.add_person(person="John Q. Hacker", note="chair")
    assert len(e.agenda[0]["related_entities"]) == 1
    e.validate()


def test_agenda_add_vote_event():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda["related_entities"] == []

    agenda.add_vote_event(vote_event="Roll no. 12")
    assert len(e.agenda[0]["related_entities"]) == 1
    e.validate()


def test_agenda_add_subject():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")

    agenda.add_subject("test")
    assert e.agenda[0]["subjects"] == ["test"]
    agenda.add_subject("test2")
    assert e.agenda[0]["subjects"] == ["test", "test2"]

    e.validate()


def test_agenda_add_classification():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")

    agenda.add_classification("test")
    assert e.agenda[0]["classification"] == ["test"]
    agenda.add_classification("test2")
    assert e.agenda[0]["classification"] == ["test", "test2"]

    e.validate()


def test_agenda_add_extra():
    e = event_obj()
    a = e.add_agenda_item("foo bar")
    a["extras"] = dict(foo=1, bar=["baz"])

    assert e.agenda[0]["extras"] == {"foo": 1, "bar": ["baz"]}


def test_add_committee():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda["related_entities"] == []

    agenda.add_committee(committee="Hello, World", note="host")
    e.validate()


def test_add_bill():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda["related_entities"] == []
    agenda.add_bill(bill="HB 101", note="consideration")
    e.validate()


def test_add_document():
    e = event_obj()
    assert e.documents == []
    e.add_document(
        note="hello",
        url="http://example.com",
        media_type="text/html",
        classification="testimony",
    )
    assert len(e.documents) == 1
    o = e.documents[0]
    assert o["note"] == "hello"
    assert o["classification"] == "testimony"
    assert o["links"] == [{"url": "http://example.com", "media_type": "text/html"}]
    e.validate()


def test_participants():
    e = event_obj()
    e.add_participant("Committee of the Whole", type="committee", note="everyone")
    assert len(e.participants) == 1
    assert e.participants[0]["name"] == "Committee of the Whole"
    assert e.participants[0]["entity_type"] == "committee"
    assert e.participants[0]["note"] == "everyone"

    # and add_person, which is a shortcut
    e.add_person("Bill Stevenson")
    assert len(e.participants) == 2
    assert e.participants[1]["name"] == "Bill Stevenson"
    assert e.participants[1]["entity_type"] == "person"
    assert e.participants[1]["note"] == "participant"


def test_set_location():
    e = event_obj()
    e.set_location(
        "North Pole",
        note="it is cold here",
        url="https://www.northpole.com",
        coordinates={"latitude": "90.0000", "longitude": "0.0000"},
    )

    assert e.location.get("name") == "North Pole"
    assert e.location.get("note") == "it is cold here"
    assert e.location.get("url") == "https://www.northpole.com"
    assert e.location.get("coordinates").get("latitude") == "90.0000"
    assert e.location.get("coordinates").get("longitude") == "0.0000"

    e.validate()


def test_add_media():
    e = event_obj()
    name = "Hello, World"
    a = e.add_agenda_item(description="foo")
    a.add_media_link(note=name, url="http://pault.ag", media_type="text/html")
    a.add_media_link(note=name, url="ftp://pault.ag", media_type="text/plain")
    e.validate()
    assert len(e.agenda[0]["media"]) == 1
    assert len(e.agenda[0]["media"][0]["links"]) == 2

    e.add_media_link(note=name, url="http://pault.ag", media_type="text/html")
    e.add_media_link(note=name, url="ftp://pault.ag", media_type="text/plain")
    e.validate()
    assert len(e.media) == 1
    assert len(e.media[0]["links"]) == 2


def test_add_bill_on_event():
    e = event_obj()
    e.add_bill("HB 6")
    e.validate()
    assert e.agenda[0]["description"] == "Associated Bills"
    assert e.agenda[0]["related_entities"][0] == {
        "bill_id": '~{"identifier": "HB 6"}',
        "entity_type": "bill",
        "name": "HB 6",
        "note": "consideration",
    }
    e.add_bill("HB 7", note="passed")
    assert e.agenda[0]["related_entities"][1] == {
        "bill_id": '~{"identifier": "HB 7"}',
        "entity_type": "bill",
        "name": "HB 7",
        "note": "passed",
    }


def test_calculate_window():
    base_day = datetime.date(2021, 9, 14)
    start, end = calculate_window(base_day=base_day)
    assert start == datetime.date(2021, 8, 15)
    assert end == datetime.date(2021, 12, 13)

    start, end = calculate_window(base_day=base_day, days_before=1, days_after=1)
    assert start == datetime.date(2021, 9, 13)
    assert end == datetime.date(2021, 9, 15)



from ..models import State, Chamber, District, simple_numbered_districts

AK = State(
    name="Alaska",
    abbr="AK",
    capital="Juneau",
    capital_tz="America/Anchorage",
    fips="02",
    unicameral=False,
    legislature_name="Alaska State Legislature",
    legislature_organization_id="ocd-organization/7971e631-d3d5-41e4-a83c-60ed2368e115",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/0aae43ba-c873-5720-a8eb-e4f03af9cc42",
    division_id="ocd-division/country:us/state:ak",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ak/government",
    url="http://w3.legis.state.ak.us/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/9c84bc85-9bb0-4f31-99ad-5fed91afdc28",
        num_seats=40,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ak", "lower", 40
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/85b8d5c1-0dde-4280-8528-9ef5d9fb5b5e",
        num_seats=20,
        title="Senator",
        districts=[
            District("A", "upper", "ocd-division/country:us/state:ak/sldu:a", 1),
            District("B", "upper", "ocd-division/country:us/state:ak/sldu:b", 1),
            District("C", "upper", "ocd-division/country:us/state:ak/sldu:c", 1),
            District("D", "upper", "ocd-division/country:us/state:ak/sldu:d", 1),
            District("E", "upper", "ocd-division/country:us/state:ak/sldu:e", 1),
            District("F", "upper", "ocd-division/country:us/state:ak/sldu:f", 1),
            District("G", "upper", "ocd-division/country:us/state:ak/sldu:g", 1),
            District("H", "upper", "ocd-division/country:us/state:ak/sldu:h", 1),
            District("I", "upper", "ocd-division/country:us/state:ak/sldu:i", 1),
            District("J", "upper", "ocd-division/country:us/state:ak/sldu:j", 1),
            District("K", "upper", "ocd-division/country:us/state:ak/sldu:k", 1),
            District("L", "upper", "ocd-division/country:us/state:ak/sldu:l", 1),
            District("M", "upper", "ocd-division/country:us/state:ak/sldu:m", 1),
            District("N", "upper", "ocd-division/country:us/state:ak/sldu:n", 1),
            District("O", "upper", "ocd-division/country:us/state:ak/sldu:o", 1),
            District("P", "upper", "ocd-division/country:us/state:ak/sldu:p", 1),
            District("Q", "upper", "ocd-division/country:us/state:ak/sldu:q", 1),
            District("R", "upper", "ocd-division/country:us/state:ak/sldu:r", 1),
            District("S", "upper", "ocd-division/country:us/state:ak/sldu:s", 1),
            District("T", "upper", "ocd-division/country:us/state:ak/sldu:t", 1),
        ],
    ),
)



from ..models import State, Chamber, simple_numbered_districts

MP = State(
    name="Northern Mariana Islands",
    abbr="MP",
    capital="Saipan",
    capital_tz="Pacific/Guam",
    fips="69",
    unicameral=False,
    # 2023-02-13 -> site is a default nginx page
    url="https://gov.mp",
    legislature_name="Northern Marianas Commonwealth Legislature",
    legislature_organization_id="ocd-organization/3266e4a2-cfa7-425a-92d3-8b18e78f0cf5",
    executive_name="Office of Governor",
    executive_organization_id="ocd-organization/568c13c8-d87b-40f0-a4a5-f1543e5f9089",
    division_id="ocd-division/country:us/territory:mp",
    jurisdiction_id="ocd-jurisdiction/country:us/territory:mp/government",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/58222d00-3b6f-422e-9b34-81350338ae91",
        num_seats=20,
        title="Legislators",
        districts=simple_numbered_districts(
            "ocd-division/country:us/territory:mp", "lower", 20
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/9f00690c-2f99-4ed0-8d0c-feadfc058054",
        num_seats=9,
        title="Senators",
        districts=simple_numbered_districts(
            "ocd-division/country:us/territory:mp", "upper", 9
        ),
    ),
)



import os
import django  # type: ignore
from django import conf  # type: ignore
import dj_database_url  # type: ignore


def init_django() -> None:  # pragma: no cover
    DATABASE_URL = os.environ.get(
        "DATABASE_URL", "postgis://openstates:openstates@localhost/openstates"
    )
    DATABASES = {"default": dj_database_url.parse(DATABASE_URL)}
    application_name = "os_core"
    if "OPTIONS" not in DATABASES:
        DATABASES["default"]["OPTIONS"] = {"application_name": application_name}
    else:
        DATABASES["default"]["OPTIONS"]["application_name"] = application_name

    try:
        conf.settings.configure(
            conf.global_settings,
            SECRET_KEY="not-important",
            DEBUG=False,
            INSTALLED_APPS=(
                "django.contrib.contenttypes",
                "openstates.data",
            ),
            DATABASES=DATABASES,
            TIME_ZONE="UTC",
            MIDDLEWARE_CLASSES=(),
        )
        django.setup()
    except RuntimeError as e:
        if "Settings already configured." not in str(e):
            raise RuntimeError(f"Encountered error {e}")



from ..models import State, Chamber, simple_numbered_districts

NJ = State(
    name="New Jersey",
    abbr="NJ",
    capital="Trenton",
    capital_tz="America/New_York",
    fips="34",
    unicameral=False,
    legislature_name="New jersey Legislature",
    legislature_organization_id="ocd-organization/8815f4a0-d799-45d1-980e-3d53e3b26b1d",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/bdfb4b37-b604-4fff-8515-e7646b6f4db6",
    division_id="ocd-division/country:us/state:nj",
    jurisdiction_id="ocd-jurisdiction/country:us/state:nj/government",
    url="http://www.njleg.state.nj.us/",
    lower=Chamber(
        chamber_type="lower",
        name="Assembly",
        organization_id="ocd-organization/68d34c4f-e59d-42a2-bcfe-fbf7f8cc8f15",
        num_seats=80,
        title="Assembly Member",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nj", "lower", 40, num_seats=2
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/0461b3ba-ecad-4ed3-bc89-056dbaa698d1",
        num_seats=40,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:nj", "upper", 40
        ),
    ),
)



import json
import typing
from .base import BaseImporter
from ._types import _JsonDict, _DBSpec, _RelatedModels
from ..utils import get_pseudo_id, _make_pseudo_id
from ..exceptions import InvalidVoteEventError
from ..data.models import VoteEvent, VoteCount, PersonVote, VoteSource, BillAction
from .organizations import OrganizationImporter
from .bills import BillImporter


class VoteEventImporter(BaseImporter):
    _type = "vote_event"
    model_class = VoteEvent
    related_models: _RelatedModels = {
        "counts": (VoteCount, "vote_event_id", {}),
        "votes": (PersonVote, "vote_event_id", {}),
        "sources": (VoteSource, "vote_event_id", {}),
    }

    def __init__(self, jurisdiction_id: str, bill_importer: BillImporter, do_postimport=True):
        super(VoteEventImporter, self).__init__(jurisdiction_id, do_postimport)
        self.org_importer = OrganizationImporter(jurisdiction_id)
        self.bill_importer = bill_importer
        self.seen_bill_ids: typing.Set[str] = set()
        self.seen_action_ids: typing.Set[str] = set()
        self.vote_events_to_delete: typing.Set[str] = set()

    def get_object(self, vote_event: _JsonDict) -> VoteEvent:
        spec = {"legislative_session_id": vote_event["legislative_session_id"]}

        if not vote_event["identifier"] and not vote_event["bill_id"]:
            raise InvalidVoteEventError(
                'attempt to save a VoteEvent without an "identifier" or "bill_id"'
            )

        if vote_event["bill_id"]:
            if vote_event["bill_id"] not in self.seen_bill_ids:
                self.seen_bill_ids.add(vote_event["bill_id"])
                # keep a list of all the vote event ids that should be deleted
                self.vote_events_to_delete.update(
                    self.model_class.objects.filter(
                        bill_id=vote_event["bill_id"]
                    ).values_list("id", flat=True)
                )
            spec["bill_id"] = vote_event["bill_id"]

        if vote_event.get("dedupe_key"):
            # replace entire spec with dedupe key
            spec = {"dedupe_key": vote_event.get("dedupe_key")}
        elif vote_event["identifier"]:
            # if there's an identifier, just use it and the bill_id and the session
            spec["identifier"] = vote_event["identifier"]
        else:
            # otherwise use the motion, start_date, and org as well
            spec.update(
                {
                    "motion_text": vote_event["motion_text"],
                    "start_date": vote_event["start_date"],
                    "organization_id": vote_event["organization_id"],
                }
            )

        return self.model_class.objects.prefetch_related("votes__voter").get(**spec)

    def limit_spec(self, spec: _JsonDict) -> _DBSpec:
        spec["legislative_session__jurisdiction_id"] = self.jurisdiction_id
        return spec

    def prepare_for_db(self, data: _JsonDict) -> _JsonDict:
        session = self.get_session(data.pop("legislative_session"))
        data["legislative_session_id"] = session.id
        organization_classification = data.pop("organization")
        data["organization_id"] = self.org_importer.resolve_json_id(
            organization_classification
        )

        bill = data.pop("bill")
        if bill and bill.startswith("~"):
            # unpack psuedo id and apply filter in case there are any that alter it
            bill = get_pseudo_id(bill)
            self.bill_importer.apply_transformers(bill)
            bill = _make_pseudo_id(**bill)

        data["bill_id"] = self.bill_importer.resolve_json_id(bill)
        bill_action = data.pop("bill_action")
        if bill_action:
            try:
                action = BillAction.objects.get(
                    bill_id=data["bill_id"],
                    description=bill_action,
                    date=data["start_date"],
                    organization_id=data["organization_id"],
                )
                # seen_action_ids is for ones being added in this import
                # action.vote is already set if action was set on prior import
                if action.id in self.seen_action_ids or hasattr(action, "vote"):
                    self.warning(
                        "can not match two VoteEvents to %s: %s", action.id, bill_action
                    )
                else:
                    data["bill_action_id"] = action.id
                    self.seen_action_ids.add(action.id)
            except BillAction.DoesNotExist:
                self.warning(
                    "could not match VoteEvent to %s %s %s",
                    bill,
                    bill_action,
                    data["start_date"],
                )
            except BillAction.MultipleObjectsReturned as e:
                self.warning(
                    "could not match VoteEvent to %s %s %s: %s",
                    bill,
                    bill_action,
                    data["start_date"],
                    e,
                )

        for vote in data["votes"]:
            vote["voter_id"] = self.resolve_person(
                vote["voter_id"], session.start_date, session.end_date,
                json.loads(organization_classification[1:]).get("classification", None)
            )
        return data

    def postimport(self) -> None:
        # be sure not to delete vote events that were imported (meaning updated) this time through
        self.vote_events_to_delete.difference_update(self.json_to_db_id.values())
        # everything remaining, goodbye
        self.model_class.objects.filter(id__in=self.vote_events_to_delete).delete()



# Generated by Django 3.2 on 2021-04-22 19:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0034_organization_other_names"),
    ]

    operations = [
        migrations.AlterField(
            model_name="bill",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="billdocument",
            name="extras",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="billversion",
            name="extras",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="event",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="eventagendaitem",
            name="extras",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="jurisdiction",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="membership",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="organization",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="organization",
            name="links",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name="organization",
            name="other_names",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name="organization",
            name="sources",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AlterField(
            model_name="person",
            name="current_role",
            field=models.JSONField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name="person",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="post",
            name="extras",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="A key-value store for storing arbitrary information not covered elsewhere.",
            ),
        ),
        migrations.AlterField(
            model_name="voteevent",
            name="extras",
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from django.db import models  # type: ignore
from .jurisdiction import Jurisdiction, LegislativeSession


OBJECT_TYPES = (
    ("jurisdiction", "Jurisdiction"),
    ("person", "Person"),
    ("organization", "Organization"),
    ("post", "Post"),
    ("membership", "Membership"),
    ("bill", "Bill"),
    ("vote_event", "VoteEvent"),
    ("event", "Event"),
)


class RunPlan(models.Model):
    jurisdiction = models.ForeignKey(
        Jurisdiction, related_name="runs", on_delete=models.CASCADE
    )
    success = models.BooleanField(default=True)
    start_time = models.DateTimeField()
    end_time = models.DateTimeField()
    exception = models.TextField(blank=True, default="")
    traceback = models.TextField(blank=True, default="")

    class Meta:
        db_table = "pupa_runplan"


class ScrapeReport(models.Model):
    plan = models.ForeignKey(RunPlan, related_name="scrapers", on_delete=models.CASCADE)
    scraper = models.CharField(max_length=300)
    args = models.CharField(max_length=300)
    start_time = models.DateTimeField()
    end_time = models.DateTimeField()

    class Meta:
        db_table = "pupa_scrapereport"


class ScrapeObjects(models.Model):
    report = models.ForeignKey(
        ScrapeReport, related_name="scraped_objects", on_delete=models.CASCADE
    )
    object_type = models.CharField(max_length=20, choices=OBJECT_TYPES)
    count = models.PositiveIntegerField()

    class Meta:
        db_table = "pupa_scrapeobjects"


class ImportObjects(models.Model):
    report = models.ForeignKey(
        RunPlan, related_name="imported_objects", on_delete=models.CASCADE
    )
    object_type = models.CharField(max_length=20, choices=OBJECT_TYPES)
    insert_count = models.PositiveIntegerField()
    update_count = models.PositiveIntegerField()
    noop_count = models.PositiveIntegerField()
    start_time = models.DateTimeField()
    end_time = models.DateTimeField()

    class Meta:
        db_table = "pupa_importobjects"


class SessionDataQualityReport(models.Model):
    legislative_session = models.ForeignKey(
        LegislativeSession, on_delete=models.CASCADE
    )

    bills_missing_actions = models.PositiveIntegerField()
    bills_missing_sponsors = models.PositiveIntegerField()
    bills_missing_versions = models.PositiveIntegerField()

    votes_missing_voters = models.PositiveIntegerField()
    votes_missing_bill = models.PositiveIntegerField()
    votes_missing_yes_count = models.PositiveIntegerField()
    votes_missing_no_count = models.PositiveIntegerField()
    votes_with_bad_counts = models.PositiveIntegerField()

    # these fields store lists of names mapped to numbers of occurances
    unmatched_sponsor_people = models.JSONField()
    unmatched_sponsor_organizations = models.JSONField()
    unmatched_voters = models.JSONField()

    class Meta:
        db_table = "pupa_sessiondataqualityreport"



from ..models import State, Chamber, District, simple_numbered_districts

ME = State(
    name="Maine",
    abbr="ME",
    capital="Augusta",
    capital_tz="America/New_York",
    fips="23",
    unicameral=False,
    legislature_name="Maine Legislature",
    legislature_organization_id="ocd-organization/1381ca6d-dedf-4bd3-8755-6f846f9e4848",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/1cb3f814-aad2-4343-8556-a2eb92f46230",
    division_id="ocd-division/country:us/state:me",
    jurisdiction_id="ocd-jurisdiction/country:us/state:me/government",
    url="http://legislature.maine.gov",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/a26314f0-dc6f-4cfc-99ca-dee4bafc9d4c",
        num_seats=152,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:me", "lower", 151
        ) + [District("Passamaquoddy Tribe", "lower",
                      "ocd-division/country:us/state:me/sldl:passamaquoddy-tribe")],
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/fe33521e-3c48-445a-adf7-75ff8a7674e4",
        num_seats=35,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:me", "upper", 35
        ),
    ),
)



import pytest  # type: ignore
from openstates.data.models import Organization, Jurisdiction, Division
from openstates.data.models import Person as DjangoPerson
from openstates.utils.people.to_database import load_person, cached_lookup
from openstates.models.people import (
    Person,
    Party,
    Link,
    Role,
    OtherName,
    OtherIdentifier,
    Office,
)
from openstates.models.committees import Committee, Membership
from openstates.cli.committees import committee_to_db, _parent_lookup


def setup():
    d = Division.objects.create(id="ocd-division/country:us/state:nc", name="NC")
    j = Jurisdiction.objects.create(
        id="ocd-jurisdiction/country:us/state:nc/government", name="NC", division=d
    )
    o = Organization.objects.create(
        name="House", classification="lower", jurisdiction=j
    )
    for n in range(1, 4):
        division = Division.objects.create(
            id=f"ocd-division/country:us/state:nc/sldl:{n}", name=str(n)
        )
        o.posts.create(label=str(n), division=division)
    Organization.objects.create(
        name="Executive", classification="executive", jurisdiction=j
    )
    Organization.objects.create(name="Democratic", classification="party")
    Organization.objects.create(name="Republican", classification="party")
    j2 = Jurisdiction.objects.create(
        id="ocd-jurisdiction/country:us/state:nc/place:cary/government", name="Cary, NC"
    )
    Organization.objects.create(
        name="Cary Town Government", classification="government", jurisdiction=j2
    )

    # clear cache here because we can't have lru_cache keep the old party ids, etc. around
    # between tests as setup() is called once per test
    cached_lookup.cache_clear()
    _parent_lookup.cache_clear()


@pytest.fixture
def person():
    PERSON_ID = "ocd-person/abcdefab-0000-1111-2222-1234567890ab"
    return Person(
        id=PERSON_ID,
        name="Jane Smith",
        party=[Party(name="Democratic")],
        roles=[],
        image="https://example.com/image",
        extras={"something": "special"},
    )


@pytest.mark.django_db
def test_basic_person_creation(person):
    created, updated = load_person(person)

    assert created is True
    p = DjangoPerson.objects.get(pk=person.id)
    assert p.name == "Jane Smith"
    assert p.image == "https://example.com/image"
    assert p.extras["something"] == "special"
    assert p.current_role is None


@pytest.mark.django_db
def test_basic_person_updates(person):
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)
    created_at, updated_at = p.created_at, p.updated_at

    # ensure no change means no change
    created, updated = load_person(person)
    assert created is False
    assert updated is False
    p = DjangoPerson.objects.get(pk=person.id)
    assert p.created_at == created_at
    assert p.updated_at == updated_at

    # ensure extra changes got captured
    person.extras["something"] = "changed"
    created, updated = load_person(person)
    assert created is False
    assert updated is True
    p = DjangoPerson.objects.get(pk=person.id)
    assert p.updated_at > updated_at
    assert p.extras["something"] == "changed"


@pytest.mark.django_db
def test_basic_person_subobjects(person):
    person.links = [
        Link(url="https://example.com"),
        Link(url="https://example.com/2", note="some additional data"),
    ]
    person.sources = [Link(url="https://example.com/jane")]
    person.other_names = [
        OtherName(name="J. Smith"),
    ]

    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    assert p.links.count() == 2
    assert p.links.filter(note="some additional data").count() == 1
    assert p.sources.count() == 1
    assert p.other_names.count() == 1


@pytest.mark.django_db
def test_subobject_update(person):
    person.links = [
        Link(url="https://example.com"),
        Link(url="https://example.com/2", note="some additional data"),
    ]
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)
    created_at, updated_at = p.created_at, p.updated_at

    # ensure no change means no change
    created, updated = load_person(person)
    assert created is False
    assert updated is False
    p = DjangoPerson.objects.get(pk=person.id)
    assert p.created_at == created_at
    assert p.updated_at == updated_at

    # change one field
    person.links[0].url = "https://example.com/jane-smith"
    created, updated = load_person(person)

    assert created is False
    assert updated is True
    p = DjangoPerson.objects.get(pk=person.id)
    assert p.links.count() == 2
    assert p.links.filter(url="https://example.com/jane-smith").count() == 1
    assert p.updated_at > updated_at

    # delete a field
    person.links.pop()
    created, updated = load_person(person)
    assert created is False
    assert updated is True
    p = DjangoPerson.objects.get(pk=person.id)
    assert p.links.count() == 1
    assert p.updated_at > updated_at


@pytest.mark.django_db
def test_subobject_duplicate(person):
    # this shouldn't actually be allowed most places (lint should catch)
    # but it was breaking committee imports when two members had the same name
    person.links = [
        Link(url="https://example.com"),
        Link(url="https://example.com"),
    ]
    # load twice, but second time no update should occur
    created, updated = load_person(person)
    created, updated = load_person(person)
    assert created is False
    assert updated is False


@pytest.mark.django_db
def test_person_identifiers(person):
    person.ids.twitter = "fakeaccount"
    person.ids.youtube = "fakeaccount"
    person.other_identifiers.append(
        OtherIdentifier(scheme="old_openstates", identifier="AR000001")
    )
    person.other_identifiers.append(
        OtherIdentifier(scheme="old_openstates", identifier="AR000002")
    )
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    assert p.identifiers.count() == 4
    assert p.identifiers.filter(scheme="old_openstates").count() == 2
    assert p.identifiers.filter(scheme="twitter")[0].identifier == "fakeaccount"


@pytest.mark.django_db
def test_person_offices(person):
    person.email = "fake@example.com"
    person.offices.append(
        Office(
            classification="capitol",
            fax="111-222-3333",
            voice="555-555-5555",
            address="123 Main St; Washington DC",
        )
    )
    person.offices.append(
        Office(
            classification="primary",
            voice="333-333-5555",
        )
    )

    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    assert p.email == "fake@example.com"
    assert p.offices.count() == 2


@pytest.mark.django_db
def test_person_party(person):
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    assert p.memberships.count() == 1
    assert p.memberships.get().organization.name == "Democratic"
    assert p.primary_party == "Democratic"

    person.party.append(Party(name="Republican", end_date="2018-10-06"))
    created, updated = load_person(person)
    assert updated is True
    assert p.primary_party == "Democratic"
    p = DjangoPerson.objects.get(pk=person.id)
    p.memberships.count() == 2
    p.memberships.exclude(end_date="").count() == 1


@pytest.mark.django_db
def test_person_legislative_roles(person):
    person.roles.append(
        Role(
            type="lower",
            district=3,
            jurisdiction="ocd-jurisdiction/country:us/state:nc/government",
        )
    )
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    # party and legislative
    assert p.memberships.count() == 2
    assert (
        p.memberships.get(organization__classification="lower").organization.name
        == "House"
    )
    assert p.memberships.get(organization__classification="lower").post.label == "3"
    assert p.current_role == {
        "org_classification": "lower",
        "district": 3,
        "division_id": "ocd-division/country:us/state:nc/sldl:3",
        "title": "Representative",
    }
    assert (
        p.current_jurisdiction_id == "ocd-jurisdiction/country:us/state:nc/government"
    )


@pytest.mark.django_db
def test_person_governor_role(person):
    person.roles.append(
        Role(
            type="governor",
            jurisdiction="ocd-jurisdiction/country:us/state:nc/government",
            end_date="2030-01-01",
        )
    )
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    assert p.memberships.count() == 2
    assert (
        p.memberships.get(organization__classification="executive").organization.name
        == "Executive"
    )
    assert p.current_role == {
        "org_classification": "executive",
        "district": None,
        "division_id": None,
        "title": "Governor",
    }
    assert (
        p.current_jurisdiction_id == "ocd-jurisdiction/country:us/state:nc/government"
    )


@pytest.mark.django_db
def test_person_mayor_role(person):
    person.roles.append(
        Role(
            type="mayor",
            jurisdiction="ocd-jurisdiction/country:us/state:nc/place:cary/government",
            end_date="2030-01-01",
        )
    )
    created, updated = load_person(person)
    p = DjangoPerson.objects.get(pk=person.id)

    assert p.memberships.count() == 2
    assert p.current_role == {
        "org_classification": "government",
        "district": None,
        "division_id": None,
        "title": "Mayor",
    }
    assert (
        p.current_jurisdiction_id
        == "ocd-jurisdiction/country:us/state:nc/place:cary/government"
    )


@pytest.mark.django_db
def test_committee_to_db_simple():
    com_id = "ocd-organization/00000000-1111-1111-1111-222222222222"
    new_com = Committee(
        id=com_id,
        name="Education",
        chamber="lower",
        jurisdiction="ocd-jurisdiction/country:us/state:nc/government",
        sources=[{"url": "https://example.com"}],
        members=[Membership(name="someone", role="member")],
    )
    created, updated, name_changed = committee_to_db(new_com)
    assert created and not updated and not name_changed
    org = Organization.objects.get(pk=com_id)
    assert org.sources == new_com.sources

    new_com.links = [{"url": "https://example.com"}]
    created, updated, name_changed = committee_to_db(new_com)
    assert updated and not created and not name_changed
    org = Organization.objects.get(pk=com_id)
    assert org.links == new_com.links

    new_com.name = "Education k-12"
    created, updated, name_changed = committee_to_db(new_com)
    assert updated and not created and name_changed
    org = Organization.objects.get(pk=com_id)
    assert org.links == new_com.links


@pytest.mark.django_db
def test_committee_to_db_memberships():
    com_id = "ocd-organization/00000000-1111-1111-1111-222222222222"
    new_com = Committee(
        id=com_id,
        name="Education",
        chamber="lower",
        jurisdiction="ocd-jurisdiction/country:us/state:nc/government",
        members=[Membership(name="Steve", role="chair")],
    )
    created, updated, name_changed = committee_to_db(new_com)
    org = Organization.objects.get(pk=com_id)
    assert org.memberships.count() == 1
    s_mem = org.memberships.get()
    assert s_mem.person_name == "Steve"
    assert s_mem.role == "chair"

    person_id = "ocd-person/33333333-4444-4444-4444-555555555555"
    wendy = DjangoPerson.objects.create(id=person_id, name="Wendy")
    new_com.add_member("Wendy", role="chair")
    new_com.members[-1].person_id = person_id
    created, updated, name_changed = committee_to_db(new_com)
    assert updated and not created and not name_changed
    org = Organization.objects.get(pk=com_id)
    assert org.memberships.count() == 2
    w_mem = org.memberships.filter(person_name="Wendy")[0]
    assert w_mem.person == wendy


@pytest.mark.django_db
def test_no_person_updates_with_committee(person):
    created, updated = load_person(person)
    DjangoPerson.objects.get(pk=person.id)

    parent_com = Committee(
        id="ocd-organization/00000000-1111-1111-1111-222222222222",
        name="Finance",
        chamber="lower",
        jurisdiction="ocd-jurisdiction/country:us/state:nc/government",
        members=[Membership(name="Steve", role="chair")],
    )
    created, updated, name_changed = committee_to_db(parent_com)

    sub_com = Committee(
        id="ocd-organization/00000000-1111-1111-1111-333333333333",
        name="Education",
        chamber="lower",
        classification="subcommittee",
        parent="ocd-organization/00000000-1111-1111-1111-222222222222",
        jurisdiction="ocd-jurisdiction/country:us/state:nc/government",
        members=[Membership(name="Steve", role="chair", person_id=person.id)],
    )
    created, updated, name_changed = committee_to_db(sub_com)
    assert created and not updated and not name_changed

    created, updated = load_person(person)
    assert not created and not updated



import datetime
import os
import typing

import requests
import logging
import shutil

from pathlib import Path

import click
import git

from ..cli.committees import to_database as committee_to_database
from ..cli.people import to_database as people_to_database

logger = logging.getLogger(__name__)


def no_recent_people_repo_commit() -> bool:
    repo_url = "https://api.github.com/repos/openstates/people/commits"
    response = requests.get(repo_url)
    no_recent_commit_exists = False
    if response.status_code == 200:
        latest_commit = response.json()[0]
        last_commit_time = latest_commit["commit"]["committer"]["date"]
        # Ensure it's UTC-aware
        last_commit_time = datetime.datetime.fromisoformat(
            last_commit_time.rstrip("Z")
        ).replace(tzinfo=datetime.timezone.utc)
        six_hours_ago = datetime.datetime.now(
            datetime.timezone.utc
        ) - datetime.timedelta(
            minutes=6 * 60 + 5
        )  # 6 hours and 5-minute buffer.
        no_recent_commit_exists = six_hours_ago > last_commit_time
    return no_recent_commit_exists


def clone_people_repo() -> None:
    repo_name = "people"
    data_path = "data"
    base_dir = (
        Path(os.environ["OS_PEOPLE_DIRECTORY"])
        if "OS_PEOPLE_DIRECTORY" in os.environ
        else Path(".")
    )
    # This is mostly for local development
    if (base_dir / data_path).is_dir():
        logger.info(f"{repo_name} directory exist")
        return

    logger.info(f"Cloning {repo_name}")
    git.refresh("/usr/bin/git")
    repo_url_http = "https://github.com/openstates/people.git"
    git.Repo.clone_from(repo_url_http, repo_name)
    logger.info(f"Done cloning {repo_name}!")

    current_directory = base_dir / repo_name
    source_data_dir = current_directory / data_path
    destination_data_dir = base_dir / data_path
    shutil.copytree(source_data_dir, destination_data_dir)


def get_args(args: list) -> list:
    def parse_value(value: str) -> typing.Union[bool, str]:
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        elif value.lower() == "none":
            return ""
        return value

    args_list = [parse_value(item) for item in args]
    return args_list


@click.group()
def main() -> None:
    pass


@main.command()
@click.argument("other-options", nargs=-1)
@click.option(
    "--purge/--no-purge",
    default=False,
    help="Set to True to purge data not in repo from database.",
)
@click.option(
    "--force-ingest/--no-force-ingest",
    default=False,
    help="Force ingest of data to database.",
)
@click.option(
    "--people/--no-people",
    default=False,
    help="Set to True to ingest only people.",
)
@click.option(
    "--committees/--no-committees",
    default=False,
    help="Set to True to ingest only committees.",
)
@click.pass_context
def update(
    ctx: typing.Any,
    other_options: list[str],
    purge: bool,
    force_ingest: bool,
    people: bool,
    committees: bool,
) -> int:
    logger.info(f"Begin ingesting people {other_options}")
    [abbr, f_ingest, pur, ppl, comms] = get_args(other_options)

    abbreviation = abbr
    purge = purge or pur
    force_ingest = force_ingest or f_ingest
    people = people or ppl
    committees = committees or comms

    if not force_ingest:
        logger.info("Checking if an update is necessary")
        if no_recent_people_repo_commit():
            logger.info(
                "There was no recent update to Openstates People Repo \n exiting..."
            )
            return 0

    logger.info(
        "An update is necessary...\nBegin updating Openstates People to-database!"
    )

    abbreviations = [abbreviation] if abbreviation else []
    clone_people_repo()
    jurisdictions_to_ingest = (
        ", ".join(abbreviations) if abbreviations else "all jurisdictions"
    )
    logger.info(
        f"Begin Openstates People Repo to Database for {jurisdictions_to_ingest}."
    )
    if people and not committees:
        ctx.invoke(people_to_database, abbreviations=abbreviations, purge=purge)
    elif committees and not people:
        ctx.invoke(committee_to_database, abbreviations=abbreviations, purge=True)
    else:
        ctx.invoke(people_to_database, abbreviations=abbreviations, purge=purge)
        ctx.invoke(committee_to_database, abbreviations=abbreviations, purge=True)
    return 0


if __name__ == "__main__":
    main()



# Generated by Django 3.0.5 on 2020-09-23 13:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0026_auto_20200923_1312"),
    ]

    operations = [
        migrations.AddField(
            model_name="eventagendamedia",
            name="classification",
            field=models.CharField(blank=True, default="", max_length=100),
        ),
    ]



import typing
import re
import tempfile
import textract  # type: ignore

from .utils import (
    pdfdata_to_text,
    text_after_line_numbers,
    text_before_line_numbers,
    text_from_element_lxml,
    text_from_element_xpath,
    text_from_element_siblings_lxml,
    text_from_element_siblings_xpath,
    clean,
)


class Metadata(typing.TypedDict):
    url: str
    media_type: str
    title: str
    jurisdiction_id: str


ExtractorFunc = typing.Callable[[bytes, Metadata], str]


def extract_simple_pdf(data: bytes, metadata: Metadata) -> str:
    return pdfdata_to_text(data)


def extract_line_numbered_pdf(data: bytes, metadata: Metadata) -> str:
    return text_after_line_numbers(pdfdata_to_text(data))


def extract_line_post_numbered_pdf(data: bytes, metadata: Metadata) -> str:
    return text_before_line_numbers(pdfdata_to_text(data))


def extract_sometimes_numbered_pdf(data: bytes, metadata: Metadata) -> str:
    """
    A few states have bills both with numbered lines and without.
    In these cases, we need to look at the start of the lines
    to determine which extraction function to use.
    """

    pdf_text = pdfdata_to_text(data)
    lines = pdf_text.split("\n")

    # Looking for lines that begin with a number
    pattern = re.compile(r"^\s*\d+\s+(.*)", flags=re.MULTILINE)
    number_of_numbered_lines = pattern.findall(pdf_text)

    # If more than 10% of the text begins with numbers, then we are
    # probably looking at a bill with numbered lines.
    THRESHOLD_NUMBERED_PDF = 0.10

    ratio_of_numbered_lines = len(number_of_numbered_lines) / len(lines)

    if ratio_of_numbered_lines > THRESHOLD_NUMBERED_PDF:
        return extract_line_numbered_pdf(data, metadata)
    else:
        return extract_simple_pdf(data, metadata)


def extract_pre_tag_html(data: bytes, metadata: Metadata) -> str:
    """
    Many states that provide bill text on HTML webpages (e.g. AK, FL)
    have the text inside <pre> tags (for preformatted text).
    """

    text_inside_matching_tag = text_from_element_lxml(data, ".//pre")
    return text_after_line_numbers(text_inside_matching_tag)


def extract_from_p_tags_html(data: bytes, metadata: Metadata) -> str:
    """
    For a few states providing bill text in HTML, we just want to get all
    the text in paragraph tags on the page. There may be several paragraphs.
    """

    text = text_from_element_siblings_lxml(data, ".//p")
    return text


def extractor_for_elements_by_class(bill_text_element_class: str) -> ExtractorFunc:
    return extractor_for_element_by_selector(
        ".//div[@class='" + bill_text_element_class + "']"
    )


def extractor_for_element_by_id(bill_text_element_id: str) -> ExtractorFunc:
    return extractor_for_element_by_selector(
        ".//div[@id='" + bill_text_element_id + "']"
    )


def extractor_for_element_by_selector(bill_text_element_selector: str) -> ExtractorFunc:
    def _my_extractor(data: bytes, metadata: Metadata) -> str:
        text_inside_matching_tag = text_from_element_lxml(
            data, bill_text_element_selector
        )
        return clean(text_inside_matching_tag)

    return _my_extractor


def extractor_for_element_by_xpath(bill_text_element_selector: str) -> ExtractorFunc:
    def _my_extractor(data: bytes, metadata: Metadata) -> str:
        text_inside_matching_tag = text_from_element_xpath(
            data, bill_text_element_selector
        )
        return clean(text_inside_matching_tag)

    return _my_extractor


def extractor_for_elements_by_xpath(bill_text_element_selector: str) -> ExtractorFunc:
    def _my_extractor(data: bytes, metadata: Metadata) -> str:
        text_inside_matching_tag = text_from_element_siblings_xpath(
            data, bill_text_element_selector
        )
        return clean(text_inside_matching_tag)

    return _my_extractor


def textract_extractor(**kwargs: str) -> ExtractorFunc:
    """ pass through kwargs to textextract.process """
    assert "extension" in kwargs, "Must supply extension"

    def func(data: bytes, metadata: Metadata) -> str:
        with tempfile.NamedTemporaryFile(delete=False) as tmpf:
            tmpf.write(data)
            tmpf.flush()
            return textract.process(tmpf.name, **kwargs).decode()

    return func


def extract_from_code_tags_html(data: bytes, metadata: Metadata) -> str:
    """
    Some states (e.g. IL) have the bill text inside
    <code> tags (as it renders as fixed-width).
    """

    text = text_from_element_siblings_lxml(data, ".//code")
    return text



import re
import datetime
import click
import typing
from pathlib import Path
from dataclasses import dataclass
from collections import defaultdict, Counter
from openstates import metadata
from enum import Enum, auto
from pydantic import ValidationError
from .retire import retire_file
from .general import (
    dump_obj,
    legacy_districts,
    load_municipalities,
)
from ...models.people import Person


class BadVacancy(Exception):
    pass


class PersonType(Enum):
    LEGISLATIVE = auto()
    RETIRED = auto()
    EXECUTIVE = auto()
    MUNICIPAL = auto()


@dataclass
class CheckResult:
    errors: list[str]
    warnings: list[str]
    fixes: list[str]


JURISDICTION_RE = re.compile(
    r"ocd-jurisdiction/country:us/(state|district|territory):\w\w/((place|county):[a-z_]+/)?government"
)

# constant to check for this particular fix
MOVED_TO_RETIRED = "moved to retired"


class Missing:
    pass


def validate_roles(
    person: Person,
    roles_key: str,
    retired: bool = False,
) -> list[str]:
    active = [role for role in getattr(person, roles_key) if role.is_active()]
    if len(active) == 0 and not retired:
        return [f"no active {roles_key}"]
    elif roles_key == "roles" and retired and len(active) > 0:
        return [f"{len(active)} active roles on retired person"]
    elif roles_key == "roles" and len(active) > 1:
        return [f"{len(active)} active roles"]
    return []


def validate_roles_key(
    person: Person,
    person_type: PersonType,
    fix: bool,
) -> CheckResult:
    resp = CheckResult([], [], [])
    role_issues = validate_roles(
        person,
        "roles",
        person_type == PersonType.RETIRED,
    )

    if person_type in (PersonType.MUNICIPAL, PersonType.EXECUTIVE) and role_issues == [
        "no active roles"
    ]:
        # municipals missing roles is a warning to avoid blocking lint
        if fix:
            resp.fixes = [MOVED_TO_RETIRED]
        else:
            resp.warnings.extend(role_issues)
    else:
        resp.errors.extend(role_issues)
    return resp


def validate_offices(person: Person) -> list[str]:
    errors = []
    type_counter: Counter[str] = Counter()
    seen_values: dict[str, str] = {}
    for office in person.offices:
        type_counter[office.classification] += 1
        for key, value in office.dict().items():
            if key == "classification" or key == "voice" or key == "fax" or not value:
                continue
            # reverse lookup to see if we've used this phone number/etc. before
            location_str = f"{office.classification} {key}"
            if value in seen_values:
                errors.append(
                    f"Value '{value}' used multiple times: {seen_values[value]} and {location_str}"
                )
            seen_values[value] = location_str
    # if type_counter["District Office"] > 1:
    #     errors.append("Multiple district offices.")
    if type_counter["capitol"] > 1:
        errors.append("Multiple capitol offices, condense to one.")
    return errors


def validate_name(person: Person, person_type: PersonType, fix: bool) -> CheckResult:
    """some basic checks on a persons name"""
    errors = []
    fixes = []
    spaces_in_name = person.name.count(" ")
    if spaces_in_name == 1:
        given_cand, family_cand = person.name.split()
        given = person.given_name
        family = person.family_name
        if not given and not family and fix:
            person.given_name = given = given_cand
            person.family_name = family = family_cand
            fixes.append(f"set given_name={given}")
            fixes.append(f"set family_name={family}")
        if not given:
            errors.append(
                f"missing given_name that could be set to '{given_cand}', run with --fix"
            )
        if not family:
            errors.append(
                f"missing family_name that could be set to '{family_cand}', run with --fix"
            )
        # expected_name = f"{given} {family}"
        # if not errors and person.data["name"] != expected_name:
        #     errors.append(f"names do not match given={given} family={family}, but name={person.data['name']}")
    return CheckResult(errors, [], fixes)


def validate_jurisdictions(person: Person, municipalities: list[str]) -> list[str]:
    errors = []
    for role in person.roles:
        try:
            metadata.lookup(jurisdiction_id=role.jurisdiction)
        except KeyError:
            if role.jurisdiction not in municipalities:
                errors.append(f"{role.jurisdiction} is not a valid jurisdiction_id")
    return errors


_EXPECTED_DISTRICTS_TYPE = dict[str, dict[str, int]]
_ACTUAL_DISTRICTS_TYPE = defaultdict[str, defaultdict[str, list[str]]]


def get_expected_districts(
    settings: dict[str, dict], abbr: str
) -> _EXPECTED_DISTRICTS_TYPE:
    expected = {}

    state = metadata.lookup(abbr=abbr)
    for chamber in state.chambers:
        chtype = (
            "legislature"
            if chamber.chamber_type == "unicameral"
            else chamber.chamber_type
        )
        expected[chtype] = {
            district.name: district.num_seats for district in chamber.districts
        }

    # remove vacancies
    vacancies = settings.get(abbr, {}).get("vacancies", [])
    if vacancies:
        click.secho(f"Processing {len(vacancies)} vacancies:")
    for vacancy in vacancies:
        if datetime.date.today() < vacancy["vacant_until"]:
            expected[vacancy["chamber"]][str(vacancy["district"])] -= 1
            click.secho(
                "\t{chamber}-{district} (until {vacant_until})".format(**vacancy),
                fg="green",
            )
        else:
            click.secho(
                "\t{chamber}-{district} expired {vacant_until} remove & re-run".format(
                    **vacancy
                ),
                fg="red",
            )
            raise BadVacancy()

    return expected


def compare_districts(
    expected: _EXPECTED_DISTRICTS_TYPE, actual: _ACTUAL_DISTRICTS_TYPE
) -> list[str]:
    errors = []

    if expected.keys() != actual.keys():
        errors.append(f"expected districts for {expected.keys()}, got {actual.keys()}")
        return errors

    for chamber in expected:
        expected_districts = set(expected[chamber].keys())
        actual_districts = set(actual[chamber].keys())
        for district in sorted(expected_districts - actual_districts):
            if expected[chamber][district]:
                errors.append(f"missing legislator for {chamber} {district}")
        for district in sorted(actual_districts - expected_districts):
            errors.append(f"extra legislator for unexpected seat {chamber} {district}")
        for district in sorted(actual_districts & expected_districts):
            if len(actual[chamber][district]) < expected[chamber][district]:
                errors.append(f"missing legislator for {chamber} {district}")
            if len(actual[chamber][district]) > expected[chamber][district]:
                people = "\n\t".join(actual[chamber][district])
                errors.append(
                    f"extra legislator for {chamber} {district}:\n\t" + people
                )
    return errors


class Validator:
    def __init__(self, abbr: str, settings: dict, fix: bool, save_all: bool):
        self.fix = fix
        self.save_all = save_all
        self.expected = get_expected_districts(settings, abbr)
        self.errors: defaultdict[str, list[str]] = defaultdict(list)
        self.warnings: defaultdict[str, list[str]] = defaultdict(list)
        self.fixes: defaultdict[str, list[str]] = defaultdict(list)
        # role type -> district -> filename
        self.active_legislators: defaultdict[
            str, defaultdict[str, list[str]]
        ] = defaultdict(lambda: defaultdict(list))
        # field name -> value -> filename
        self.duplicate_values: defaultdict[
            str, defaultdict[str, list[str]]
        ] = defaultdict(lambda: defaultdict(list))
        self.legacy_districts = legacy_districts(abbr=abbr)
        self.municipalities = [m["id"] for m in load_municipalities(abbr=abbr)]
        for m in self.municipalities:
            if not JURISDICTION_RE.match(m):
                raise ValueError(f"invalid municipality id {m}")

    def process_validator_result(
        self,
        validator_func: typing.Callable[[Person, PersonType, bool], CheckResult],
        person: Person,
        person_type: PersonType,
        original_filename: Path,
    ) -> None:
        result = validator_func(person, person_type, self.fix)
        self.errors[original_filename.name].extend(result.errors)
        self.warnings[original_filename.name].extend(result.warnings)
        if result.fixes:
            self.fixes[original_filename.name].extend(result.fixes)
            dump_obj(person, filename=original_filename)

    def validate_person(
        self,
        data: dict[str, typing.Any],
        filename: Path,
        person_type: PersonType,
    ) -> None:
        print_filename = filename.name
        try:
            person = Person(**data)
            self.errors[print_filename] = []
        except ValidationError as ve:
            self.errors[print_filename] = [
                f"  {'.'.join(str(l) for l in error['loc'])}: {error['msg']}"
                for error in ve.errors()
            ]
            # if we couldn't create a valid person, bail now
            return

        uid = person.id.split("/")[1]
        if uid not in print_filename:
            self.errors[print_filename].append(f"id piece {uid} not in filename")

        self.errors[print_filename].extend(
            validate_jurisdictions(person, self.municipalities)
        )

        # looser validation for upstream-maintained unitedstates.io data
        if "/us/legislature" not in str(filename):
            self.errors[print_filename].extend(validate_offices(person))

        self.process_validator_result(validate_roles_key, person, person_type, filename)
        self.process_validator_result(validate_name, person, person_type, filename)

        if person_type == PersonType.RETIRED:
            self.errors[print_filename].extend(self.validate_old_district_names(person))

        # check duplicate IDs
        self.duplicate_values["openstates"][person.id].append(print_filename)
        for scheme, value in person.ids.dict().items():
            if value:
                self.duplicate_values[scheme][value].append(print_filename)
        for ident in person.other_identifiers:
            self.duplicate_values[ident.scheme][ident.identifier].append(print_filename)

        # update active legislators
        if person_type == PersonType.LEGISLATIVE:
            role_type = district = None
            for role in person.roles:
                if role.is_active():
                    role_type = role.type
                    district = role.district
                    break
            self.active_legislators[str(role_type)][str(district)].append(
                print_filename
            )

        # special case for the auto-retirement fix
        if MOVED_TO_RETIRED in self.fixes[print_filename]:
            retire_file(filename)

        if self.save_all:
            # save person if save_all, will re-order fields/etc.
            click.secho(f"re-saving {filename}", fg="green")
            dump_obj(person, filename=filename)

    def validate_old_district_names(self, person: Person) -> list[str]:
        errors = []
        for role in person.roles:
            if (
                role.district
                and role.district not in self.expected[role.type]
                and role.district not in self.legacy_districts[role.type]
            ):
                errors.append(f"unknown district name: {role.type} {role.district}")
        return errors

    def check_duplicates(self) -> list[str]:
        """
        duplicates should already be stored in self.duplicate_values
        this method just needs to turn them into errors
        """
        errors = []
        for key, values in self.duplicate_values.items():
            for value, instances in values.items():
                if len(instances) > 1:
                    if len(instances) > 3:
                        instance_str = ", ".join(instances[:3])
                        instance_str += " and {} more...".format(len(instances) - 3)
                    else:
                        instance_str = ", ".join(instances)
                    errors.append(f'duplicate {key}: "{value}" {instance_str}')
        return errors

    def print_validation_report(self, verbose: bool) -> int:  # pragma: no cover
        error_count = 0

        for fn, errors in self.errors.items():
            warnings = self.warnings[fn]
            fixes = self.fixes[fn]
            if errors or warnings or fixes:
                click.echo(fn)
                for fix in fixes:
                    click.secho(" " + fix, fg="green")
                for err in errors:
                    click.secho(" " + err, fg="red")
                    error_count += 1
                for warning in warnings:
                    click.secho(" " + warning, fg="yellow")
            if not errors and verbose > 0:
                click.secho(fn + " OK!", fg="green")

        for err in self.check_duplicates():
            click.secho(err, fg="red")
            error_count += 1

        errors = compare_districts(self.expected, self.active_legislators)
        for err in errors:
            click.secho(err, fg="red")
            error_count += 1

        return error_count



import re
import sys
import json
import uuid
import typing
from functools import lru_cache
from pathlib import Path
from enum import Enum
from dataclasses import dataclass
from collections import defaultdict
import click
import yaml
from django.db import transaction  # type: ignore
from yaml.representer import Representer
from pydantic import ValidationError
from ..metadata import lookup
from ..utils.django import init_django  # type: ignore
from ..utils.people import get_data_path, get_all_abbreviations
from ..utils.people.to_database import CancelTransaction
from ..models.committees import Committee, ScrapeCommittee
from ..models.people import Person

yaml.SafeDumper.add_representer(defaultdict, Representer.represent_dict)
yaml.SafeDumper.add_multi_representer(Enum, Representer.represent_str)  # type: ignore


@dataclass
class DirectoryMergePlan:
    names_to_add: set[str]
    names_to_remove: set[str]
    same: int
    to_merge: list[tuple[Committee, ScrapeCommittee]]


class PersonMatcher:
    def __init__(self, abbr: str, directory: typing.Optional[Path] = None):
        self.abbr = abbr
        # chamber -> name piece -> id set
        self.current_people: dict[str, dict[str, set[str]]] = {
            "upper": {},
            "lower": {},
            "legislature": {},
        }
        self.all_ids: set[str] = set()

        # allow directory override for testing purposes
        if not directory:
            directory = get_data_path(abbr) / "legislature"

        # read in people with current roles
        for filename in directory.glob("*.yml"):
            person: Person = Person.load_yaml(filename)
            chamber = ""
            for role in person.roles:
                if role.is_active():
                    chamber = role.type
                    break
            self.add_name(chamber, person.name, person.id)
            if person.family_name:
                self.add_name(chamber, person.family_name, person.id)
            for name in person.other_names:
                self.add_name(chamber, name.name, person.id)

    def add_name(self, chamber: str, name_piece: str, id_: str) -> None:
        name_piece = name_piece.lower()
        self.all_ids.add(id_)
        if name_piece in self.current_people[chamber]:
            self.current_people[chamber][name_piece].add(id_)
        else:
            self.current_people[chamber][name_piece] = {id_}
        # add to legislature too
        if chamber != "legislature":
            self.add_name("legislature", name_piece, id_)

    @lru_cache(500)
    def match(self, chamber: str, name: str) -> typing.Optional[str]:
        name = name.lower()
        candidates = self.current_people[chamber].get(name, None)
        if not candidates:
            click.secho(
                f"  no candidates while attempting to match {chamber} {name}",
                fg="yellow",
            )
            return None
        elif len(candidates) == 1:
            return list(candidates)[0]
        else:
            click.secho(
                f"  multiple candidates while attempting to match {chamber} {name}",
                fg="yellow",
            )
            return None

    def id_exists(self, id_: str) -> bool:
        return id_ in self.all_ids


@lru_cache(5)
def _parent_lookup(jurisdiction_id: str, chamber: str, parent: str) -> str:
    from openstates.data.models import Organization

    if parent:
        return parent
    else:
        return Organization.objects.get(
            jurisdiction_id=jurisdiction_id, classification=chamber
        ).id


def committee_to_db(com: Committee) -> tuple[bool, bool, bool]:
    from openstates.data.models import Organization, Membership

    updated = False
    name_changed = False

    db_com, created = Organization.objects.get_or_create(
        id=com.id,
        jurisdiction_id=com.jurisdiction,
        parent_id=_parent_lookup(com.jurisdiction, com.chamber, com.parent),
        classification=com.classification,
        defaults=dict(name=com.name, extras=com.extras),
    )

    if db_com.name != com.name:
        updated = True
        name_changed = True
        db_com.name = com.name

    if db_com.extras != com.extras:
        updated = True
        db_com.extras = com.extras

    for key_name in ("links", "sources", "other_names"):
        list_json = [n.dict() for n in getattr(com, key_name)]
        if list_json != getattr(db_com, key_name):
            setattr(db_com, key_name, list_json)
            updated = True

    existing_members = {
        (m.person_id, m.person_name, m.role)
        for m in db_com.memberships.all().select_related("person")
    }
    scraped_members = {(m.person_id, m.name, m.role) for m in com.members}

    if existing_members != scraped_members:
        # replace members
        updated = True
        new_memberships = [
            Membership(
                role=m.role,
                person_name=m.name,
                person_id=m.person_id,
                organization=db_com,
            )
            for m in com.members
        ]
        db_com.memberships.all().delete()
        Membership.objects.bulk_create(new_memberships)

    if updated:
        db_com.save()
        # don't set updated to true in return if created
        if created:
            updated = False
    return created, updated, name_changed


def merge_lists(orig: list, new: list, key_attr: str) -> list:
    """merge two lists based on a unique property"""
    combined = []
    new_by_key = {getattr(item, key_attr): item for item in new}
    seen = set()
    # add original items, or their replacements if present
    for item in orig:
        key = getattr(item, key_attr)
        if key in seen:
            continue
        seen.add(key)
        if key in new_by_key:
            combined.append(new_by_key[key])
        else:
            combined.append(item)
    # add new items
    for key, item in new_by_key.items():
        if key not in seen:
            combined.append(item)
    return combined


class CommitteeDir:
    def __init__(
        self,
        abbr: str,
        raise_errors: bool = True,
        directory: typing.Optional[Path] = None,
    ):
        self.abbr = abbr
        # allow overriding directory explicitly, useful for testing
        self.directory = directory if directory else get_data_path(abbr) / "committees"
        # chamber -> name -> Committee
        self.coms_by_parent_and_name: defaultdict[str, dict[str, Committee]] = (
            defaultdict(dict)
        )
        self.errors = []
        # person matcher will be prepared if/when needed
        self.person_matcher: typing.Optional[PersonMatcher] = None

        # make sure a committees dir exists
        self.directory.mkdir(parents=True, exist_ok=True)

        for filename in self.directory.glob("*.yml"):
            try:
                com: Committee = Committee.load_yaml(filename)
                self.coms_by_parent_and_name[com.parent or com.chamber][com.name] = com
            except ValidationError as ve:
                if raise_errors:
                    raise
                self.errors.append((filename, ve))

    def merge_committees(self, orig: Committee, new: ScrapeCommittee) -> Committee:
        # need new parent id
        new_parent_id = None
        if new.parent:
            new_parent_id = self.coms_by_parent_and_name[new.chamber][new.parent].id

        # disallow merge of these, likely error & unclear what should happen
        if orig.chamber != new.chamber:
            raise ValueError("cannot merge committees with different chambers")
        if orig.parent != new_parent_id:
            print(orig)
            print(new_parent_id)
            print(new)
            raise ValueError("cannot merge committees with different parents")
        if orig.classification != new.classification:
            raise ValueError("cannot merge committees with different classifications")
        # TODO: jurisdiction isn't yet set on ScrapeCommittee... do we need another check here?

        merged = Committee(
            id=orig.id,  # id stays constant
            chamber=orig.chamber,
            parent=orig.parent,
            classification=orig.classification,
            jurisdiction=orig.jurisdiction,
            name=new.name,  # name can be updated
            sources=merge_lists(orig.sources, new.sources, "url"),
            links=merge_lists(orig.links, new.links, "url"),
            other_names=merge_lists(orig.other_names, new.other_names, "name"),
            members=merge_lists(orig.members, new.members, "name"),
            extras=orig.extras | new.extras,
        )
        return merged

    def print_warnings(self) -> None:
        unmatched_names = set()
        for coms_for_chamber in self.coms_by_parent_and_name.values():
            for com in coms_for_chamber.values():
                has_homepage = "homepage" in [link.note for link in com.links]
                if com.classification == "committee" and not has_homepage:
                    click.secho(f"{com.name} does not have homepage link", fg="yellow")
                for membership in com.members:
                    if not membership.person_id:
                        unmatched_names.add(membership.name)

        if unmatched_names:
            click.secho(f"{len(unmatched_names)} unmatched legislators:", fg="yellow")
            for name in unmatched_names:
                click.secho(f"    {name}", fg="yellow")
            click.secho("try running with --fix", fg="yellow")

    def get_new_filename(self, obj: Committee) -> str:
        id = obj.id.split("/")[1]
        name = re.sub(r"\s+", "-", obj.name)
        name = re.sub(r"[^a-zA-Z-]", "", name)
        return f"{obj.chamber}-{name}-{id}.yml"

    def get_filename_by_id(self, com_id: str) -> Path:
        if com_id.startswith("ocd-organization"):
            com_id = com_id.split("/")[1]
        assert len(com_id) == 36
        files = list(self.directory.glob(f"*{com_id}.yml"))
        if len(files) == 1:
            return files[0]
        else:
            raise FileNotFoundError()

    def get_filename_by_name(self, parent: str, name: str) -> Path:
        try:
            com = self.coms_by_parent_and_name[parent][name]
        except KeyError:
            raise FileNotFoundError()
        return self.get_filename_by_id(com.id)

    def save_committee(self, committee: Committee) -> None:
        # try to use id's existing filename if possible
        try:
            filename = self.get_filename_by_id(committee.id)
        except FileNotFoundError:
            filename = self.directory / self.get_new_filename(committee)

        with open(filename, "w") as f:
            yaml.dump(
                committee.to_dict(),
                f,
                default_flow_style=False,
                Dumper=yaml.SafeDumper,
                sort_keys=False,
            )

    def add_committee(self, committee: ScrapeCommittee) -> None:
        parent_id: typing.Optional[str]
        # parent name needs to be converted to ID
        if committee.parent:
            parent_id = self.coms_by_parent_and_name[committee.chamber][
                committee.parent
            ].id
        else:
            parent_id = None
        # convert a ScrapeCommittee to a committee by giving it an ID
        full_com = Committee(
            id=f"ocd-organization/{uuid.uuid4()}",
            jurisdiction=lookup(abbr=self.abbr).jurisdiction_id,
            parent=parent_id,
            **committee.dict(exclude={"parent"}),
        )
        self.coms_by_parent_and_name[committee.parent or committee.chamber][
            committee.name
        ] = full_com
        self.save_committee(full_com)

    def ingest_scraped_json(self, input_dir: str) -> list[ScrapeCommittee]:
        if not self.person_matcher:
            self.person_matcher = PersonMatcher(self.abbr)

        scraped_data = []
        for filename in Path(input_dir).glob("*"):
            with open(filename) as file:
                data = json.load(file)
                if not data:
                    click.secho(f"no committee in {filename}", fg="yellow")
                    continue
                com = ScrapeCommittee(**data)
                # do person matching on import so that diffs work
                for member in com.members:
                    mid = self.person_matcher.match(com.chamber, member.name)
                    if mid:
                        member.person_id = mid
                scraped_data.append(com)
        return scraped_data

    def update_unmatched_names(self) -> None:
        if not self.person_matcher:
            self.person_matcher = PersonMatcher(self.abbr)

        # find all committees with unmatched names
        for coms_for_chamber in self.coms_by_parent_and_name.values():
            for com in coms_for_chamber.values():
                updated_count = 0
                for membership in com.members:
                    if not membership.person_id:
                        # do the matching
                        mid = self.person_matcher.match(com.chamber, membership.name)
                        if mid:
                            membership.person_id = mid
                            updated_count += 1
                if updated_count:
                    self.save_committee(com)
                    click.secho(
                        f"updated {updated_count} memberships on {com.name} with names",
                        fg="yellow",
                    )

    def get_merge_plan_by_parent(
        self, parent: str, new_data: list[ScrapeCommittee]
    ) -> DirectoryMergePlan:
        existing_names = set(self.coms_by_parent_and_name[parent].keys())
        new_names = {com.name for com in new_data}
        parent_names = {
            c.id: c.name for c in self.coms_by_parent_and_name[parent].values()
        }

        names_to_add = new_names - existing_names
        names_to_remove = existing_names - new_names
        names_to_compare = new_names & existing_names
        to_merge = list()
        same = 0

        for com in new_data:
            if com.name in names_to_compare:
                # reverse a saved Committee to a ScrapeCommittee for comparison
                existing = self.coms_by_parent_and_name[parent][com.name]
                com_without_id = existing.dict()
                if com_without_id["parent"]:
                    com_without_id["parent"] = parent_names[com_without_id["parent"]]
                com_without_id.pop("id")
                com_without_id.pop("jurisdiction")
                rev_sc = ScrapeCommittee(**com_without_id)
                if com != rev_sc and com.parent == rev_sc.parent:
                    to_merge.append((existing, com))
                else:
                    same += 1

        return DirectoryMergePlan(
            names_to_add=names_to_add,
            names_to_remove=names_to_remove,
            same=same,
            to_merge=to_merge,
        )

    def to_database(self, purge: bool) -> None:
        from openstates.data.models import Organization

        ids = set()
        created_count = 0
        updated_count = 0
        name_change_count = 0

        jurisdiction_id = lookup(abbr=self.abbr).jurisdiction_id
        existing_ids = set(
            Organization.objects.filter(
                jurisdiction_id=jurisdiction_id,
                classification__in=("committee", "subcommittee"),
            ).values_list("id", flat=True)
        )

        for parent, committees in self.coms_by_parent_and_name.items():
            # this sorted hack ensures subcommittees are processed after all committees
            for name, committee in sorted(
                committees.items(), key=lambda c: c[1].parent or ""
            ):
                ids.add(committee.id)
                created, updated, name_changed = committee_to_db(committee)

                if created:
                    click.secho(f"created committee {name}", fg="cyan", bold=True)
                    created_count += 1
                elif updated:
                    click.secho(f"updated committee {name}", fg="cyan")
                    updated_count += 1

                if name_changed:
                    name_change_count += 1

        missing_ids = existing_ids - ids

        # ids that are missing need to be purged
        if missing_ids and not purge:
            click.secho(
                f"{len(missing_ids)} went missing, run with --purge to remove", fg="red"
            )
            for id in missing_ids:
                mobj = Organization.objects.get(pk=id)
                click.secho(f"  {id}: {mobj}")
            raise CancelTransaction()
        elif missing_ids and purge:
            click.secho(f"{len(missing_ids)} purged", fg="yellow")
            Organization.objects.filter(id__in=missing_ids).delete()

        click.secho(
            f"processed {len(ids)} committees, {created_count} created, "
            f"{updated_count} updated"
            f" {name_change_count} name changed",
            fg="green",
        )


@click.group()
def main() -> None:
    pass


@main.command()  # pragma: no cover
@click.argument("abbr")
@click.argument("input_dir")
@click.option("--interactive/--no-interactive", default=False)
def merge(abbr: str, input_dir: str, interactive: bool) -> None:
    """
    Merge scraped committee data into repo.
    """
    comdir = CommitteeDir(abbr)

    coms_by_parent: defaultdict[str, list[ScrapeCommittee]] = defaultdict(list)
    scraped_data = comdir.ingest_scraped_json(input_dir)
    for com in scraped_data:
        coms_by_parent[com.parent or com.chamber].append(com)

    for parent, coms in coms_by_parent.items():
        plan = comdir.get_merge_plan_by_parent(parent, coms)

        click.secho(
            f"{len(plan.names_to_add)} to add",
            fg="yellow" if plan.names_to_add else "green",
        )
        click.secho(
            f"{len(plan.names_to_remove)} to remove",
            fg="yellow" if plan.names_to_remove else "green",
        )
        click.secho(f"{plan.same} without changes", fg="green")
        click.secho(
            f"{len(plan.to_merge)} with changes",
            fg="yellow" if plan.to_merge else "green",
        )

        if plan.names_to_add or plan.names_to_remove or plan.to_merge:
            if interactive and not click.confirm("Do you wish to continue?"):
                sys.exit(1)

            # add new committees
            # sort key ensures parents are added before children
            for com in sorted(coms, key=lambda c: c.parent or ""):
                if com.name in plan.names_to_add:
                    comdir.add_committee(com)
                    click.secho(f"  adding {com.chamber} {com.name}")

            # remove old committees
            for name in plan.names_to_remove:
                filename = comdir.get_filename_by_name(parent, name)
                click.secho(f"removing {filename}", fg="red")
                filename.unlink()

            # merge remaining committees
            for orig, new in plan.to_merge:
                merged = comdir.merge_committees(orig, new)
                comdir.save_committee(merged)
        else:
            click.secho("nothing to do!", fg="green")


@main.command()  # pragma: no cover
@click.argument("abbreviations", nargs=-1)
@click.option(
    "--fix/--no-fix", default=False, help="Enable/disable automatic fixing of data."
)
def lint(abbreviations: list[str], fix: bool) -> None:
    """
    Lint committee YAML files.
    """
    if not abbreviations:
        abbreviations = get_all_abbreviations()

    for abbr in abbreviations:
        comdir = CommitteeDir(abbr, raise_errors=False)
        errors = 0
        click.secho(f"==== {abbr} ====")
        for filename, error in comdir.errors:
            click.secho(filename.name)
            for err in error.errors():
                click.secho(
                    f"  {'.'.join(str(l) for l in err['loc'])}: {err['msg']}", fg="red"
                )
                errors += 1
        if fix:
            comdir.update_unmatched_names()
        comdir.print_warnings()
        if errors:
            click.secho(f"exiting with {errors} errors", fg="red")
            sys.exit(1)


@main.command()
@click.argument("abbreviations", nargs=-1)
@click.option(
    "--purge/--no-purge",
    default=False,
    help="Purge all committees from DB that aren't in YAML.",
)
@click.option(
    "--safe/--no-safe",
    default=False,
    help="Operate in safe mode, no changes will be written to database.",
)
def to_database(abbreviations: list[str], purge: bool, safe: bool) -> None:
    """
    Sync YAML files to DB.
    """
    init_django()

    if not abbreviations:
        abbreviations = get_all_abbreviations()

    for abbr in abbreviations:
        click.secho("==== {} ====".format(abbr), bold=True)
        comdir = CommitteeDir(abbr)

        if safe:
            click.secho("running in safe mode, no changes will be made", fg="magenta")

        try:
            with transaction.atomic():
                comdir.to_database(purge=purge)
                if safe:
                    click.secho("ran in safe mode, no changes were made", fg="magenta")
                    raise CancelTransaction()
        except CancelTransaction:
            sys.exit(1)


if __name__ == "__main__":
    main()



import pytest
from openstates.scrape import State
from openstates.importers import JurisdictionImporter
from openstates.data.models import Jurisdiction, Division, LegislativeSession


class NewJersey(State):
    legislative_sessions = [
        {"identifier": "2015", "name": "2015 Regular Session"},
        {"identifier": "2016", "name": "2016 Regular Session"},
    ]


@pytest.mark.django_db
def test_jurisdiction_import():
    Division.objects.create(id="ocd-division/country:us/state:nj", name="NJ")
    tj = NewJersey()
    juris_dict = tj.as_dict()
    JurisdictionImporter("jurisdiction-id").import_data([juris_dict])

    dbj = Jurisdiction.objects.get()
    assert dbj.id == tj.jurisdiction_id
    assert dbj.division_id == tj.division_id
    assert dbj.name == tj.name
    assert dbj.url == tj.url


@pytest.mark.django_db
def test_jurisdiction_update():
    Division.objects.create(id="ocd-division/country:us/state:nj", name="NJ")
    tj = NewJersey()
    ji = JurisdictionImporter("jurisdiction-id")
    _, what = ji.import_item(tj.as_dict())
    assert what == "insert"

    _, what = ji.import_item(tj.as_dict())
    assert what == "noop"
    assert Jurisdiction.objects.count() == 1

    tj.extras = {"something": "here"}
    obj, what = ji.import_item(tj.as_dict())
    assert what == "update"
    assert Jurisdiction.objects.count() == 1
    assert Jurisdiction.objects.get().extras == {"something": "here"}


@pytest.mark.django_db
def test_jurisdiction_merge_related():
    Division.objects.create(id="ocd-division/country:us/state:nj", name="NJ")
    # need to ensure legislative_sessions don't get deleted
    ji = JurisdictionImporter("jurisdiction-id")
    tj = NewJersey()
    ji.import_item(tj.as_dict())

    assert LegislativeSession.objects.count() == 2

    # disallow deletion of legislative sessions as it can remove bills
    tj.legislative_sessions.pop()
    ji.import_item(tj.as_dict())

    # should still have two
    assert LegislativeSession.objects.count() == 2

    # now will have three
    tj.legislative_sessions.append({"identifier": "2017", "name": "2017 Session"})
    ji.import_item(tj.as_dict())
    assert LegislativeSession.objects.count() == 3

    # and test that the non-identifier fields actually update
    tj.legislative_sessions.append({"identifier": "2016", "name": "updated"})
    ji.import_item(tj.as_dict())
    assert LegislativeSession.objects.count() == 3
    assert LegislativeSession.objects.get(identifier="2016").name == "updated"



from ..models import State, Chamber, simple_numbered_districts

KS = State(
    name="Kansas",
    abbr="KS",
    capital="Topeka",
    capital_tz="America/Chicago",
    fips="20",
    unicameral=False,
    legislature_name="Kansas State Legislature",
    legislature_organization_id="ocd-organization/6ee5506b-e33f-45c3-965e-59a531f249e6",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/433be248-0fd6-54cd-883b-2a1ab5f8b45d",
    division_id="ocd-division/country:us/state:ks",
    jurisdiction_id="ocd-jurisdiction/country:us/state:ks/government",
    url="http://www.kslegislature.org/",
    lower=Chamber(
        chamber_type="lower",
        name="House",
        organization_id="ocd-organization/31d2e44e-f94c-4971-83ea-e1fe29613c73",
        num_seats=125,
        title="Representative",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ks", "lower", 125
        ),
    ),
    upper=Chamber(
        chamber_type="upper",
        name="Senate",
        organization_id="ocd-organization/c1d465f9-fab5-4cc0-aded-270423bf31ba",
        num_seats=40,
        title="Senator",
        districts=simple_numbered_districts(
            "ocd-division/country:us/state:ks", "upper", 40
        ),
    ),
)



# Generated by Django 3.0.5 on 2020-04-22 00:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("data", "0003_auto_20200422_0031")]

    operations = [
        migrations.RemoveField(model_name="membership", name="on_behalf_of"),
        migrations.RemoveField(model_name="person", name="national_identity"),
        migrations.RemoveField(model_name="person", name="sort_name"),
    ]
